__all__ = ()

from scarletio.http_client import ConnectorTCP

from ..core import KOKORO


CONNECTOR = None


def get_connector():
    """
    Gets tcp connector. If already retrieved once returns that instead.
    
    Returns
    -------
    connector : ``ConnectorTCP`` 
    """
    global CONNECTOR
    if (CONNECTOR is None) or CONNECTOR.closed:
        CONNECTOR = ConnectorTCP(KOKORO)
    
    return CONNECTOR



__all__ = ()

from ...channel import Channel, create_partial_channel_from_data
from ...field_validators import nullable_entity_dictionary_validator_factory
from ...message import Attachment, Message
from ...role import Role
from ...user import ClientUserBase, User

# attachments

def parse_attachments(data):
    """
    Parsers out the resolved attachments from the given data.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Resolved data.
    
    Returns
    -------
    attachments : `None`, `dict` of (`int`, ``Attachment``) items
    """
    resolved_attachment_datas = data.get('attachments', None)
    if (resolved_attachment_datas is None) or (not resolved_attachment_datas):
        return
    
    resolved_attachments = {}
    
    for attachment_data in resolved_attachment_datas.values():
        attachment = Attachment.from_data(attachment_data)
        resolved_attachments[attachment.id] = attachment
    
    return resolved_attachments


def put_attachments(attachments, data, defaults):
    """
    Puts the given `attachments` into the given `data` json serializable object.
    
    Parameters
    ----------
    attachments : `None`, `dict` of (`int`, ``Attachment``) items
        Resolved attachments.
    
    data : `dict<str, object>`
        Interaction resolved data.
    
    defaults : `bool`
        Whether default fields values should be included as well.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    if defaults or (attachments is not None):
        resolved_attachment_datas = {}
        
        if (attachments is not None):
            for attachment in attachments.values():
                resolved_attachment_datas[str(attachment.id)] = attachment.to_data(
                    defaults = defaults, include_internals = True
                )
        
        data['attachments'] = resolved_attachment_datas
    
    return data


validate_attachments = nullable_entity_dictionary_validator_factory('attachments', Attachment)

# channels

def parse_channels(data, guild_id = 0):
    """
    Parsers out the resolved channels from the given data.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Resolved data.
    
    guild_id : `int`, Optional (Keyword only)
        The respective guild's identifier.
    
    Returns
    -------
    channels : `None`, `dict` of (`int`, ``Channel``) items
    """
    resolved_channel_datas = data.get('channels', None)
    if (resolved_channel_datas is None) or (not resolved_channel_datas):
        return
    
    resolved_channels = {}
    
    for channel_data in resolved_channel_datas.values():
        channel = create_partial_channel_from_data(channel_data, guild_id)
        if (channel is not None):
            resolved_channels[channel.id] = channel
    
    return resolved_channels


def put_channels(channels, data, defaults):
    """
    Puts the given `channels` into the given `data` json serializable object.
    
    Parameters
    ----------
    channels : `None`, `dict` of (`int`, ``Channel``) items
        Resolved channels.
    
    data : `dict<str, object>`
        Interaction resolved data.
    
    defaults : `bool`
        Whether default fields values should be included as well.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    if defaults or (channels is not None):
        resolved_channel_datas = {}
        
        if (channels is not None):
            for channel in channels.values():
                resolved_channel_datas[str(channel.id)] = channel.to_data(
                    defaults = defaults, include_internals = True
                )
        
        data['channels'] = resolved_channel_datas
    
    return data


validate_channels = nullable_entity_dictionary_validator_factory('channels', Channel)

# roles

def parse_roles(data, guild_id = 0):
    """
    Parsers out the resolved roles from the given data.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Resolved data.
    
    guild_id : `int`, Optional (Keyword only)
        The respective guild's identifier.
    
    Returns
    -------
    roles : `None`, `dict` of (`int`, ``Role``) items
    """
    resolved_role_datas = data.get('roles', None)
    if (resolved_role_datas is None) or (not resolved_role_datas):
        return
    
    resolved_roles = {}
    
    for role_data in resolved_role_datas.values():
        role = Role.from_data(role_data, guild_id)
        resolved_roles[role.id] = role
    
    return resolved_roles


def put_roles(roles, data, defaults):
    """
    Puts the given `roles` into the given `data` json serializable object.
    
    Parameters
    ----------
    roles : `None`, `dict` of (`int`, ``Role``) items
        Resolved roles.
    
    data : `dict<str, object>`
        Interaction resolved data.
    
    defaults : `bool`
        Whether default fields values should be included as well.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    if defaults or (roles is not None):
        resolved_role_datas = {}
        
        if (roles is not None):
            for role in roles.values():
                resolved_role_datas[str(role.id)] = role.to_data(
                    defaults = defaults, include_internals = True
                )
        
        data['roles'] = resolved_role_datas
    
    return data


validate_roles = nullable_entity_dictionary_validator_factory('roles', Role)

# messages

def parse_messages(data):
    """
    Parsers out the resolved messages from the given data.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Resolved data.
    
    Returns
    -------
    messages : `None`, `dict` of (`int`, ``Message``) items
    """
    resolved_message_datas = data.get('messages', None)
    if (resolved_message_datas is None) or (not resolved_message_datas):
        return
    
    resolved_messages = {}
    
    for message_data in resolved_message_datas.values():
        message = Message.from_data(message_data)
        resolved_messages[message.id] = message
    
    return resolved_messages


def put_messages(messages, data, defaults):
    """
    Puts the given `messages` into the given `data` json serializable object.
    
    Parameters
    ----------
    messages : `None`, `dict` of (`int`, ``Message``) items
        Resolved messages.
    
    data : `dict<str, object>`
        Interaction resolved data.
    
    defaults : `bool`
        Whether default fields values should be included as well.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    if defaults or (messages is not None):
        resolved_message_datas = {}
        
        if (messages is not None):
            for message in messages.values():
                resolved_message_datas[str(message.id)] = message.to_data(
                    defaults = defaults, include_internals = True
                )
        
        data['messages'] = resolved_message_datas
    
    return data


validate_messages = nullable_entity_dictionary_validator_factory('messages', Message)

# users

def parse_users(data, guild_id = 0):
    """
    Parsers out the resolved users from the given data.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Resolved data.
    
    guild_id : `int` = `0`, Optional
        The received interaction's guild's identifier.
    
    Returns
    -------
    messages : `None`, `dict` of (`int`, ``ClientUserBase``) items
    """
    resolved_user_datas = data.get('users', None)
    if (resolved_user_datas is None) or (not resolved_user_datas):
        return
    
    resolved_guild_profile_datas = data.get('members', None)

    resolved_users = {}
    
    for user_id, user_data in resolved_user_datas.items():
        
        if resolved_guild_profile_datas is None:
            guild_profile_data = None
        else:
            guild_profile_data = resolved_guild_profile_datas.get(user_id, None)
        
        user = User.from_data(user_data, guild_profile_data, guild_id)
        resolved_users[user.id] = user
    
    return resolved_users


def put_users(users, data, defaults, *, guild_id = 0):
    """
    Puts the given `users` into the given `data` json serializable object.
    
    Parameters
    ----------
    users : `None`, `dict` of (`int`, ``ClientUserBase``) items
        Resolved users.
    
    data : `dict<str, object>`
        Interaction resolved data.
    
    defaults : `bool`
        Whether default fields values should be included as well.
    
    guild_id : `int` = `0`, Optional (Keyword only)
        The respective guild's identifier to use for handing user guild profiles.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    if defaults or (users is not None):
        resolved_user_datas = {}
        resolved_guild_profile_datas = {}
        
        if (users is not None):
            for user in users.values():
                resolved_user_datas[str(user.id)] = user.to_data(
                    defaults = defaults, include_internals = True
                )
                
                guild_profile = user.guild_profiles.get(guild_id, None)
                if (guild_profile is not None):
                    resolved_guild_profile_datas[str(user.id)] = guild_profile.to_data(
                        defaults = defaults, include_internals = True
                    )
        
        data['users'] = resolved_user_datas
        data['members'] = resolved_guild_profile_datas
    
    return data


validate_users = nullable_entity_dictionary_validator_factory('users', ClientUserBase)



__all__ = ('ClientWrapper',)

from scarletio import RichAttributeErrorBaseType

from ..core import CLIENTS

from .client import Client


def optimize_keyword_parameters(keyword_parameters):
    """
    Optimizes the given keyword parameters omitting ellipses.
    
    Parameters
    ----------
    keyword_parameters : `dict<str, object>`
        Keyword parameters to optimize.
    
    Returns
    -------
    keyword_parameters : `None | dict<str, object>`
    """
    keyword_parameters = {key: value for key, value in keyword_parameters.items() if value is not ...}
    if not keyword_parameters:
        keyword_parameters = None
    
    return keyword_parameters


class ClientWrapperEventsProxy(RichAttributeErrorBaseType):
    """
    When a client wrapper's `.events` is called without giving the `func` parameter to it an instance
    of this class is created for allowing using it as a decorator with passing additional keyword parameters at the
    same time.
    
    Attributes
    ----------
    called : `bool`
        Whether the poxy was already called and should fail the next time if `func` is not given.
    client_wrapper : ``ClientWrapper``
        The owner event descriptor.
    keyword_parameters: `None | dict<str, object>`
        Additional keyword parameters passed when the wrapper was created.
    """
    __slots__ = ('called', 'client_wrapper', 'keyword_parameters')
    
    def __new__(cls, client_wrapper, called, keyword_parameters):
        """
        Creates an instance from the given parameters.
        
        Parameters
        ----------
        client_wrapper : ``ClientWrapper``
            The owner client wrapper.
        keyword_parameters: `None | dict<str, object>`
            Additional keyword parameters passed when the wrapper was created.
        """
        self = object.__new__(cls)
        cls.called.__set__(self, called)
        cls.client_wrapper.__set__(self, client_wrapper)
        cls.keyword_parameters.__set__(self, keyword_parameters)
        return self
    
    
    def __call__(self, func = None, *, name = ..., overwrite = ...):
        """
        Adds the given `func` as event handler to the client_wrapper's clients' with the stored up parameters.
        
        Parameters
        ----------
        func : `callable`
            The event handler to add to the respective clients.
        
        Returns
        -------
        func : `None`, `callable` = `None`, Optional
            The event handler to add to the respective clients.
            
            If not given, will return a decorator.
        
        name : `None`, `str`, Optional (Keyword only)
            The name to which event the handler should be registered to.
            
            If not given, it will be extracted from the handler's name.
        
        overwrite : `bool`, Optional (Keyword only)
            Whether the current event handler(s) should be replaced by the added one.
        
        Returns
        -------
        output : `instance<type<self>>` / `func`
            The given an events proxy if `func` is `None` or omitted.
        
        Raises
        ------
        AttributeError
            - Invalid event name.
        TypeError
            - If a parameter's type is incorrect.
        """
        if self.called:
            if func is None:
                raise TypeError('`func` cannot be given as `None`.')
            
            keyword_parameters = self.keyword_parameters
        else:
            keyword_parameters = optimize_keyword_parameters({'name': name, 'overwrite': overwrite})
            
            if func is None:
                return type(self)(self.client_wrapper, True, keyword_parameters)
        
        for client in self.client_wrapper.clients:
            event_handler = client.events
            if keyword_parameters is None:
                event_handler(func)
            else:
                event_handler(func, **keyword_parameters)
        
        return func
        
    
    def __repr__(self):
        """Returns the proxy's representation."""
        repr_parts = ['<', self.__class__.__name__]
        
        repr_parts.append(' client_wrapper = ')
        repr_parts.append(repr(self.client_wrapper))
        
        called = self.called
        if called:
            repr_parts.append(', called = ')
            repr_parts.append(repr(called))
        
        keyword_parameters = self.keyword_parameters
        if (keyword_parameters is not None):
            repr_parts.append(', keyword_parameters = ')
            repr_parts.append(repr(keyword_parameters))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __eq__(self, other):
        """Returns whether the proxys are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        if self.called != other.called:
            return False
        
        if self.client_wrapper != other.client_wrapper:
            return False
        
        if self.keyword_parameters != other.keyword_parameters:
            return False
        
        return True
    
    
    def __setattr__(self, attribute_name, attribute_value):
        """
        Sets an attribute of the wrapped client's event handler managers.
        
        Parameters
        ----------
        attribute_name : `str`
            The name of the event.
        attribute_value : `callable`
            The event handler.
        
        Raises
        ------
        AttributeError
            ``EventHandlerManager`` has no attribute named as the given `attribute_name`.
        """
        for client in self.client_wrapper.clients:
            client.events.__setattr__(attribute_name, attribute_value)
    
    
    def __delattr__(self, attribute_name):
        """
        Deletes an attribute of the wrapped client's event handler managers.
        
        Parameters
        ----------
        attribute_name : `str`
            The name of the event.
        
        Raises
        ------
        AttributeError
            ``EventHandlerManager`` has no attribute named as the given `attribute_name`.
        """
        for client in self.client_wrapper.clients:
            client.events.__delattr__(attribute_name)


class ClientWrapper(RichAttributeErrorBaseType):
    """
    Wraps together more clients enabling to add the same event handlers or commands to them. Tho for that feature, you
    need first import it's respective extension.
    
    Attributes
    ----------
    clients : ``Clients``
        The clients to wrap together.
    """
    __slots__ = ('clients',)
    
    def __new__(cls, *clients):
        """
        Creates a new ``ClientWrapper`` with the given clients. If no clients are given, then will wrap
        all the clients.
        
        Parameters
        ----------
        *clients : ``Client``
            The clients to wrap together.
        
        Raises
        ------
        TypeError
            A non ``Client`` was given.
        """
        if clients:
            for client in clients:
                if not isinstance(client, Client):
                    raise TypeError(
                        f'{cls.__name__} expects only `{Client.__name__}`, '
                        f'got {client.__class__.__name__}; {client!r}.'
                    )
        else:
            clients = tuple(CLIENTS.values())
        
        self = object.__new__(cls)
        self.clients = clients
        return self
    
    
    def __repr__(self):
        """Returns the client wrapper's representation."""
        result = [self.__class__.__name__, '(']
        
        clients = self.clients
        limit = len(clients)
        if limit:
            index = 0
            while True:
                client = clients[index]
                result.append(client.full_name)
                
                index += 1
                if index == limit:
                    break
                
                result.append(', ')
                continue
        
        result.append(')')
        
        return ''.join(result)
    
    
    def __eq__(self, other):
        if type(other) is not type(self):
            return NotImplemented
        
        return {*self.clients} == {*other.clients}
    
    
    def __iter__(self):
        """Iterates over the clients of the client wrapper."""
        yield from self.clients
    
    
    def __contains__(self, value):
        """Returns whether the client wrapper contains the given value."""
        return value in self.clients
    
    
    @property
    def events(self):
        """
        Returns an events proxy which enables adding / removing event handlers.
        
        Returns
        -------
        proxy : ``ClientWrapperEventsProxy``
        """
        return ClientWrapperEventsProxy(self, False, None)



__all__ = ()
from scarletio import to_json

from ...discord.utils import datetime_to_timestamp

from .constants import (
    JSON_KEY_BOT_INFO_AVATAR_BASE64, JSON_KEY_BOT_INFO_BANNER_URL, JSON_KEY_BOT_INFO_CERTIFIED_AT,
    JSON_KEY_BOT_INFO_DISCRIMINATOR_STRING, JSON_KEY_BOT_INFO_DONATE_BOT_GUILD_ID,
    JSON_KEY_BOT_INFO_FEATURED_GUILD_ID_ARRAY, JSON_KEY_BOT_INFO_GITHUB_URL, JSON_KEY_BOT_INFO_ID,
    JSON_KEY_BOT_INFO_INVITE_URL, JSON_KEY_BOT_INFO_IS_CERTIFIED, JSON_KEY_BOT_INFO_LONG_DESCRIPTION,
    JSON_KEY_BOT_INFO_NAME, JSON_KEY_BOT_INFO_OWNER_ID_ARRAY, JSON_KEY_BOT_INFO_PREFIX,
    JSON_KEY_BOT_INFO_SHORT_DESCRIPTION, JSON_KEY_BOT_INFO_SUPPORT_SERVER_INVITE_URL, JSON_KEY_BOT_INFO_TAG_ARRAY,
    JSON_KEY_BOT_INFO_UPVOTES, JSON_KEY_BOT_INFO_UPVOTES_MONTHLY, JSON_KEY_BOT_INFO_VANITY_URL,
    JSON_KEY_BOT_INFO_WEBSITE_URL
)


def snowflake_to_string(snowflake):
    """
    Converts snowflake value to string.
    
    Parameters
    ----------
    snowflake : `int`
        Snowflake to convert.
    
    Returns
    -------
    snowflake : `int`
    """
    return int(snowflake)


def discriminator_to_string(discriminator):
    """
    Converts a discriminator back to string.
    
    Parameters
    ----------
    discriminator : `int`
        The discriminator to convert.
    
    Returns
    -------
    discriminator : `str`
    """
    return str(discriminator)


def avatar_to_base16(avatar):
    """
    Converts a avatar to base16.
    
    Parameters
    ----------
    avatar : ``Icon``
        The icon to convert.
    
    Returns
    -------
    avatar : `str`
    """
    return avatar.as_base_16_hash


def nullable_string_tuple_to_string_array(nullable_string_tuple):
    """
    Converts a nullable string tuple to an array representing it.
    
    Parameters
    ----------
    nullable_string_tuple : `None | tuple<str>`
        The value to convert.
    
    Returns
    -------
    array : `list` of `str`
    """
    if nullable_string_tuple is None:
        array = []
    else:
        array = list(nullable_string_tuple)
    
    return array


def nullable_snowflake_tuple_to_string_array(nullable_snowflake_tuple):
    """
    Converts a nullable snowflake tuple to an array representing it.
    
    Parameters
    ----------
    nullable_snowflake_tuple : `None | tuple<int>`
        The value to convert.
    
    Returns
    -------
    array : `list` of `str`
    """
    if nullable_snowflake_tuple is None:
        array = []
    else:
        array = [str(snowflake) for snowflake in nullable_snowflake_tuple]
    
    return array


def owner_id_to_string_array(owner_id):
    """
    Converts a snowflake owner id to an array of strings.
    
    Parameters
    ----------
    owner_id : `int`
        Snowflake value.
    
    Returns
    -------
    owner_id_array : `list` of `str`
    """
    return [str(owner_id)]


def owner_ids_to_string_array(owner_ids):
    """
    Converts a snowflake owner id to an array of strings.
    
    Parameters
    ----------
    owner_ids : `iterable` of `int`
        Snowflakes.
    
    Returns
    -------
    owner_id_array : `list` of `str`
    """
    return [str(owner_id) for owner_id in owner_ids]


def nullable_datetime_to_timestamp(date_time):
    """
    Converts a nullable datetime to timestamp.
    
    Parameters
    ----------
    date_time : `None | DateTime`
        The date time to convert.
    
    Returns
    -------
    timestamp : `None`, `str`
    """
    if date_time is None:
        timestamp = None
    else:
        timestamp = datetime_to_timestamp(date_time)
    
    return timestamp


BOT_INFO_ATTRIBUTE_TO_FIELD_NAME = {
    'id': JSON_KEY_BOT_INFO_ID,
    'name': JSON_KEY_BOT_INFO_NAME,
    'discriminator': JSON_KEY_BOT_INFO_DISCRIMINATOR_STRING,
    'avatar': JSON_KEY_BOT_INFO_AVATAR_BASE64,
    'banner_url': JSON_KEY_BOT_INFO_BANNER_URL,
    'prefix': JSON_KEY_BOT_INFO_PREFIX,
    'short_description': JSON_KEY_BOT_INFO_SHORT_DESCRIPTION,
    'long_description': JSON_KEY_BOT_INFO_LONG_DESCRIPTION,
    'tags': JSON_KEY_BOT_INFO_TAG_ARRAY,
    'website_url': JSON_KEY_BOT_INFO_WEBSITE_URL,
    'support_server_invite_url': JSON_KEY_BOT_INFO_SUPPORT_SERVER_INVITE_URL,
    'github_url': JSON_KEY_BOT_INFO_GITHUB_URL,
    'owner_id': JSON_KEY_BOT_INFO_OWNER_ID_ARRAY,
    'owner_ids': JSON_KEY_BOT_INFO_OWNER_ID_ARRAY,
    'featured_guild_ids': JSON_KEY_BOT_INFO_FEATURED_GUILD_ID_ARRAY,
    'invite_url': JSON_KEY_BOT_INFO_INVITE_URL,
    'certified_at': JSON_KEY_BOT_INFO_CERTIFIED_AT,
    'is_certified': JSON_KEY_BOT_INFO_IS_CERTIFIED,
    'vanity_url': JSON_KEY_BOT_INFO_VANITY_URL,
    'upvotes': JSON_KEY_BOT_INFO_UPVOTES,
    'upvotes_monthly': JSON_KEY_BOT_INFO_UPVOTES_MONTHLY,
    'donate_bot_guild_id': JSON_KEY_BOT_INFO_DONATE_BOT_GUILD_ID,
}

BOT_INFO_ATTRIBUTE_TO_FIELD_NAME_AND_CONVERTER = {
    'id': (JSON_KEY_BOT_INFO_ID, snowflake_to_string),
    'name': (JSON_KEY_BOT_INFO_NAME, None),
    'discriminator': (JSON_KEY_BOT_INFO_DISCRIMINATOR_STRING, discriminator_to_string),
    'avatar': (JSON_KEY_BOT_INFO_AVATAR_BASE64, avatar_to_base16),
    'banner_url': (JSON_KEY_BOT_INFO_BANNER_URL, None),
    'prefix': (JSON_KEY_BOT_INFO_PREFIX, None),
    'short_description': (JSON_KEY_BOT_INFO_SHORT_DESCRIPTION, None),
    'long_description': (JSON_KEY_BOT_INFO_LONG_DESCRIPTION, None),
    'tags': (JSON_KEY_BOT_INFO_TAG_ARRAY, None),
    'website_url': (JSON_KEY_BOT_INFO_WEBSITE_URL, None),
    'support_server_invite_url': (JSON_KEY_BOT_INFO_SUPPORT_SERVER_INVITE_URL, None),
    'github_url': (JSON_KEY_BOT_INFO_GITHUB_URL, None),
    'owner_id': (JSON_KEY_BOT_INFO_OWNER_ID_ARRAY, owner_id_to_string_array),
    'owner_ids': (JSON_KEY_BOT_INFO_OWNER_ID_ARRAY, owner_ids_to_string_array),
    'featured_guild_ids': (JSON_KEY_BOT_INFO_FEATURED_GUILD_ID_ARRAY, nullable_snowflake_tuple_to_string_array),
    'invite_url': (JSON_KEY_BOT_INFO_INVITE_URL, None),
    'certified_at': (JSON_KEY_BOT_INFO_CERTIFIED_AT, nullable_datetime_to_timestamp),
    'is_certified': (JSON_KEY_BOT_INFO_IS_CERTIFIED, None),
    'vanity_url': (JSON_KEY_BOT_INFO_VANITY_URL, None),
    'upvotes': (JSON_KEY_BOT_INFO_UPVOTES, None),
    'upvotes_monthly': (JSON_KEY_BOT_INFO_UPVOTES_MONTHLY, None),
    'donate_bot_guild_id': (JSON_KEY_BOT_INFO_DONATE_BOT_GUILD_ID, snowflake_to_string),
}


# used when generating exception messages.
FIELD_NAMES_STRING = ', '.join(
    repr(field_name) for field_name in BOT_INFO_ATTRIBUTE_TO_FIELD_NAME_AND_CONVERTER.keys()
)

def get_bots_query_sort_by_value(sort_by):
    """
    Gets bot query sort by field.
    
    Parameters
    ----------
    sort_by : `None`, `str`
        Sort by parameter.
    
    Returns
    -------
    query_sort_by_value : `str`
        The built query value.
    
    Raises
    ------
    LookupError
        Unknown field name given.
    """
    if (sort_by is None) or (not sort_by):
        query_sort_by_value = ''
    else:
        try:
            query_sort_by_value = BOT_INFO_ATTRIBUTE_TO_FIELD_NAME[sort_by]
        except KeyError:
            raise LookupError(
                f'There is no query field named: {sort_by!r}. The accepted fields are the '
                f'following: {FIELD_NAMES_STRING}.'
            ) from None
    
    return query_sort_by_value


def create_bots_query_search_value(query):
    """
    Creates a bots query search value.
    
    Parameters
    ----------
    query : `None | dict<str, object>`
        Search parameters.
    
    Returns
    -------
    query_search_value : `str`
        The built query value.
    
    Raises
    ------
    LookupError
        Unknown field name given.
    """
    if (query is None) or (not query):
        query_search_value = ''
    else:
        query_parts = []
        
        for field_name, field_value in query.items():
            try:
                top_gg_name, converter = BOT_INFO_ATTRIBUTE_TO_FIELD_NAME[field_name]
            except KeyError:
                raise LookupError(
                    f'There is no query field named: {field_name!r}. The accepted fields are the '
                    f'following: {FIELD_NAMES_STRING}.'
                ) from None
            
            if converter is None:
                top_gg_value = field_value
            else:
                top_gg_value = converter(field_value)
            
            top_gg_value = to_json(top_gg_value)
            query_parts.append((top_gg_name, top_gg_value))
        
        limit = len(query_parts)
        if limit:
            string_parts = []
            
            index = 0
            
            while True:
                top_gg_name, top_gg_value = query_parts[index]
                string_parts.append(top_gg_name)
                string_parts.append(': ')
                string_parts.append(top_gg_value)
                
                if index == limit:
                    break
                
                string_parts.append(' ')
                continue
            
            query_search_value = ''.join(string_parts)
            
        else:
            query_search_value = ''
    
    return query_search_value


BOTS_QUERY_FIELDS_VALUE = ', '.join((
    JSON_KEY_BOT_INFO_ID,
    JSON_KEY_BOT_INFO_NAME,
    JSON_KEY_BOT_INFO_DISCRIMINATOR_STRING,
    JSON_KEY_BOT_INFO_AVATAR_BASE64,
    JSON_KEY_BOT_INFO_BANNER_URL,
    JSON_KEY_BOT_INFO_PREFIX,
    JSON_KEY_BOT_INFO_SHORT_DESCRIPTION,
    JSON_KEY_BOT_INFO_LONG_DESCRIPTION,
    JSON_KEY_BOT_INFO_TAG_ARRAY,
    JSON_KEY_BOT_INFO_WEBSITE_URL,
    JSON_KEY_BOT_INFO_SUPPORT_SERVER_INVITE_URL,
    JSON_KEY_BOT_INFO_GITHUB_URL,
    JSON_KEY_BOT_INFO_OWNER_ID_ARRAY,
    JSON_KEY_BOT_INFO_FEATURED_GUILD_ID_ARRAY,
    JSON_KEY_BOT_INFO_INVITE_URL,
    JSON_KEY_BOT_INFO_CERTIFIED_AT,
    JSON_KEY_BOT_INFO_IS_CERTIFIED,
    JSON_KEY_BOT_INFO_VANITY_URL,
    JSON_KEY_BOT_INFO_UPVOTES,
    JSON_KEY_BOT_INFO_UPVOTES_MONTHLY,
    JSON_KEY_BOT_INFO_DONATE_BOT_GUILD_ID,
))



__all__ = ('create_ansi_format_code',)


def create_ansi_format_code(text_decoration = None, background_color = None, foreground_color = None):
    """
    Creates an ansi text format code for `ansi` codeblocks.
    
    > If no parameter is given, will generate a format reset code.
    
    Parameters
    ----------
    text_decoration : `None`, ``AnsiTextDecoration`` = `None`, Optional
        Text decoration.
    
    background_color : `None`, ``AnsiBackgroundColor`` = `None`, Optional
        background color.
    
    foreground_color : `None`, ``AnsiForegroundColor`` = `None`, Optional
        Foreground color.
    
    Returns
    -------
    format_code : `str`
    """
    format_code_parts = ['\u001b[']
    
    field_added = False
    
    if (text_decoration is not None):
        field_added = True
        
        format_code_parts.append(str(text_decoration.value))
    
    if (background_color is not None):
        if field_added:
            format_code_parts.append(';')
        else:
            field_added = True
        
        format_code_parts.append(str(background_color.value))
    
    if (foreground_color is not None):
        if field_added:
            format_code_parts.append(';')
        else:
            field_added = True
        
        format_code_parts.append(str(foreground_color.value))
    
    if not field_added:
        format_code_parts.append('0')
    
    format_code_parts.append('m')
    
    return ''.join(format_code_parts)



__all__ = ('call_command', 'execute_command_from_system_parameters',)

import sys

from ...discord import KOKORO

from .command import CommandResult, normalize_command_name
from .command.result import (
    COMMAND_RESULT_CODE_COMMAND_NOT_AVAILABLE,  COMMAND_RESULT_CODE_COMMAND_NOT_FOUND,
    COMMAND_RESULT_CODE_COMMAND_REQUIRED
)
from .constants import REGISTERED_COMMANDS_BY_NAME, SYSTEM_DEFAULT_PARAMETER
from .lookup import maybe_find_commands


def call_command(parameters, index, output_stream):
    """
    Tries to call a command.
    
    Parameters
    ----------
    parameters : `list` of `str`
        Command line parameters.
    index : `int`
        The index of the first parameter trying to process.
    output_stream : `stream-like`
        Output stream.
    """
    maybe_find_commands()
    
    # Use goto
    while True:
        if index >= len(parameters):
            command_result = CommandResult(
                COMMAND_RESULT_CODE_COMMAND_REQUIRED,
            )
            break
    
        command_name = parameters[index]
        index += 1
        command_name = normalize_command_name(command_name)
        
        try:
            command = REGISTERED_COMMANDS_BY_NAME[command_name]
        except KeyError:
            command_result = CommandResult(
                COMMAND_RESULT_CODE_COMMAND_NOT_FOUND,
                command_name,
            )
            break
            
        if not command.available:
            command_result = CommandResult(
                COMMAND_RESULT_CODE_COMMAND_NOT_AVAILABLE,
                command_name,
            )
            break
        
        command_result = command.invoke(parameters, index)
        break
            
    
    output = command_result.get_message()
    if (output is not None):
        output_stream.write(output)
        
        if not output.endswith('\n'):
            output_stream.write('\n')
    


def execute_command_from_system_parameters():
    """
    Calls the respective command from system parameters.
    """
    system_parameters = sys.argv
    if len(system_parameters) < 2:
        system_parameters = [*system_parameters, SYSTEM_DEFAULT_PARAMETER]
    
    try:
        call_command(system_parameters, 1, sys.stdout)
    except BaseException as err:
        exception = err
    else:
        exception = None
    
    # Stop the event loop on exception or if we are doing nothing.
    if KOKORO.running and (exception is not None) or (not KOKORO.get_tasks()):
        KOKORO.stop()
    
    if (exception is not None):
        raise exception



__all__ = ('ReactionMappingLine',)

from scarletio import RichAttributeErrorBaseType

from .fields import validate_count, validate_users


class ReactionMappingLine(RichAttributeErrorBaseType):
    """
    Contains who reaction with a represented reaction.
    
    Attributes
    ----------
    count : `int`
        The total number of users who reacted.
    users : ``None | set<ClientUserBase>``
        The known reactors.
    """
    __slots__ = ('count', 'users')
    
    def __new__(cls, *, count = ..., users = ...):
        # count
        if count is ...:
            count = 0
        else:
            count = validate_count(count)
        
        # users
        if users is ...:
            users = None
        else:
            users = validate_users(users)
        
        # postprocess
        if (users is not None) and (len(users) > count):
            count = len(users)
        
        # Construct
        self = object.__new__(cls)
        self.count = count
        self.users = users
        return self
    
    
    def __len__(self):
        """Returns the amount of known users."""
        users = self.users
        if users is None:
            return 0
        
        return len(users)
    
    
    def __iter__(self):
        """Iterates over the known users."""
        users = self.users
        if (users is not None):
            yield from users
    
    
    def __contains__(self, user):
        """Returns whether the given user is in the known ones."""
        users = self.users
        if (users is None):
            return False
        
        return user in users
    
    
    def __repr__(self):
        """Returns the representation of the container."""
        repr_parts = ['<', type(self).__name__]
        
        # count
        repr_parts.append(', count = ')
        repr_parts.append(repr(self.count))
        
        # users
        repr_parts.append(', users = ')
        repr_parts.append(repr(self.users))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __eq__(self, other):
        """Returns whether the two poll results are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        # count
        if self.count != other.count:
            return False
        
        # users
        length = len(self)
        if length != len(other) or (length != 0 and self.users != other.users):
            return False
        
        return True
    
    
    def __hash__(self):
        """Returns the count's hash."""
        hash_value = 0
        
        # count
        hash_value ^= self.count
        
        # users
        users = self.users
        if (users is not None):
            hash_value ^= len(users) << 4
            for user in users:
                hash_value ^= hash(user)
        
        return hash_value
    
    
    def _merge_with(self, other):
        """
        Merges the poll result with the other one.
        
        Always `other` is right and `other` should be empty.
        
        Parameters
        ----------
        other : `instance<type<self>>`
            Other instance.
        """
        new_count = other.count
        
        self_users = self.users
        other_users = other.users
        if self_users is None:
            if (other_users is not None):
                self_users = other_users.copy()
                self.users = self_users
        
        else:
            if self.count != new_count:
                self_users.clear()
                if (other_users is not None):
                    self_users.update(other_users)
            
            else:
                if (other_users is not None) and (not other_users.issubset(self_users)):
                    self_users.clear()
                    self_users.update(other_users)
            
        self.count = new_count
    
    
    @classmethod
    def _create_empty(cls):
        """
        Creates a new reaction mapping line.
        
        Parameters
        ----------
        count : `int`
            The amount of reactions to create with.
        
        Returns
        -------
        self : instance<cls>
        """
        self = object.__new__(cls)
        self.count = 0
        self.users = None
        return self
    
    
    def copy(self):
        """
        Copies the reaction mapping line.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        
        new.count = self.count
        
        users = self.users
        if (users is not None):
            users = users.copy()
        new.users = users
        
        return new
    
    
    def copy_with(self, *, count = ..., users = ...):
        """
        Copies the reaction mapping line with the given fields.
        
        Parameters
        ----------
        count : `int`, Optional (Keyword only)
            The amount of reactors.
        
        users : `None`, `iterable` of ``ClientUserBase``, Optional (Keyword only)
            The known users.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # count
        if count is ...:
            count = self.count
        else:
            count = validate_count(count)
        
        # users
        if users is ...:
            users = self.users
            if (users is not None):
                users = users.copy()
        else:
            users = validate_users(users)
        
        # post-checks
        if (users is not None) and (len(users) > count):
            count = len(users)
        
        # Construct
        new = object.__new__(type(self))
        new.count = count
        new.users = users
        return new
    
    
    def _add_reaction(self, user):
        """
        Adds a user.
        
        Parameters
        ----------
        user : ``ClientUserBase``
            The user who reacted.
        
        Returns
        -------
        success : `bool`
        """
        users = self.users
        if users is None:
            self.users = {user}
            self.count += 1
            return True
        
        old_length = len(users)
        users.add(user)
        new_length = len(users)
        if old_length == new_length:
            return False
        
        self.count += 1
        return True
    
    
    def _fill_some_reactions(self, some_users):
        """
        Fills out some users.
        
        Parameters
        ----------
        some_users : ``list<ClientUserBase>``
            The users who reacted.
        """
        users = self.users
        if users is None:
            if some_users:
                users = {*some_users}
                self.users = users
        else:
            users.update(some_users)
        
        # This should not happen
        if (users is not None) and (self.count < len(users)):
            self.count = len(users)
    
    
    def _fill_all_reactions(self, all_users):
        """
        Fills out all users.
        
        Parameters
        ----------
        all_users : ``list<ClientUserBase>``
            The users who reacted.
        """
        users = self.users
        if users is None:
            if all_users:
                users = {*all_users}
                self.users = users
        else:
            users.clear()
            users.update(all_users)
        
        # This should not happen
        user_count = 0 if users is None else len(users)
        if self.count != user_count:
            self.count = user_count
    
    
    def _remove_reaction(self, user):
        """
        Removes a user.
        
        Parameters
        ----------
        user : ``ClientUserBase``
            The user who removed their reaction.
        
        Returns
        -------
        success : `bool`
        """
        users = self.users
        count = self.count
        
        if users is None:
            if count <= 0:
                return False
            
            self.count = count - 1
            return True
        
        old_length = len(users)
        users.discard(user)
        new_length = len(users)
        if old_length == new_length:
            count = self.count
            if count <= new_length:
                return False
            
            self.count = count - 1
            return True
        
        if count <= 0:
            return False
        
        self.count = count - 1
        return True
    
    
    @property
    def unknown(self):
        """
        Returns the amount of unknown users.
        
        Returns
        -------
        unknown : `int`
        """
        count = self.count
        users = self.users
        if users is None:
            return count
        
        return count - len(users)
    
    
    def filter_after(self, limit, after):
        """
        If we know all the users, then instead of executing a Discord API request we filter the users locally
        using this method.
        
        Parameters
        ----------
        limit : `int`
            The maximal limit of the users to return.
        after : `int`
            Gets the users after this specified id.
        
        Returns
        -------
        users : ``list<ClientUserBase>``
        """
        users = self.users
        if users is None:
            return []
        
        filtered_users = [user for user in self if user.id > after]
        filtered_users.sort()
        del filtered_users[limit:]
        return filtered_users
    
    
    def clear(self):
        """
        Clears the reaction mapping line's users.
        """
        users = self.users
        if (users is not None):
            users.clear()



__all__ = ('GuildJoinRequestFormResponse',)

from scarletio import RichAttributeErrorBaseType


class GuildJoinRequestFormResponse(RichAttributeErrorBaseType):
    """
    Form response of a guild join request.
    
    > No known fields.
    """
    def __new__(cls):
        """
        Creates a new guild join request form response with the given fields.
        
        > No known fields.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        self = object.__new__(cls)
        return self
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new guild join request form response instance from the given data.
        
        Parameters
        ----------
        data : `dict<str, object>`.
            Form response data.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        return self
    
    
    def to_data(self, *, defaults = False):
        """
        Converts the guild join request form response into a json serializable object.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether default field values should be included as well.
        
        Returns
        -------
        data : `dict<str, object>`
        """
        data = {}
        return data
    
    
    def __repr__(self):
        """Returns the guild join request form response's representation."""
        repr_parts = ['<', self.__class__.__name__]
        
        # nothing yet
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __eq__(self, other):
        """Returns whether the two join requests responses are the same."""
        if type(self) is not type(other):
            return NotImplemented
        
        # Nothing yet
        
        return True
    
    
    def __hash__(self):
        """Returns the guild join request form response's hash value."""
        return 0
    
    
    def copy(self):
        """
        Copies the guild join request form response.
        
        Returns
        -------
        new : `instance<type<cls>>`
        """
        new = object.__new__(type(self))
        return new
    
    
    def copy_with(self):
        """
        Copies the guild join request form response with the given data.
        
        > No known fields.
        
        Returns
        -------
        new : `instance<type<cls>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        new = object.__new__(type(self))
        return new



__all__ = ('ScheduledEventOccasionOverwriteUpdateEvent',)

from ..scheduled_event_occasion_overwrite_create_event import ScheduledEventOccasionOverwriteCreateEvent


class ScheduledEventOccasionOverwriteUpdateEvent(ScheduledEventOccasionOverwriteCreateEvent):
    """
    Represents a `GUILD_SCHEDULED_EVENT_EXCEPTION_CREATE` event where the occasion overwrite already exists.
    
    Attributes
    ----------
    guild_id : `int`
        The guild's identifier where the event is for.
    
    occasion_overwrite : ``ScheduledEventOccasionOverwrite``
        The affected occasion overwrite.
    
    scheduled_event_id : `int`
        The scheduled event's identifier.
    """
    __slots__ = ()



__all__ = ()

from ...field_parsers import nullable_string_parser_factory
from ...field_putters import nullable_string_putter_factory
from ...field_validators import nullable_string_validator_factory

from .constants import TEXT_LENGTH_MAX


# text

parse_text = nullable_string_parser_factory('text')
put_text = nullable_string_putter_factory('text')
validate_text = nullable_string_validator_factory('text', 0, TEXT_LENGTH_MAX)



__all__ = ()

from scarletio import copy_docs, include

from ....discord.application_command import ApplicationCommandOption, ApplicationCommandOptionChoice

from ..constants import APPLICATION_COMMAND_FUNCTION_DEEPNESS
from ..converter_constants import (
    ANNOTATION_AUTO_COMPLETE_AVAILABILITY, ANNOTATION_TYPE_TO_OPTION_TYPE, ANNOTATION_TYPE_TO_REPRESENTATION,
    ANNOTATION_TYPE_TO_STR_ANNOTATION
)
from ..exceptions import SlashCommandParameterConversionError

from .base import ParameterConverterBase


SlashCommandParameterAutoCompleter = include('SlashCommandParameterAutoCompleter')


class ParameterConverterSlashCommand(ParameterConverterBase):
    """
    Converter class for slash command options.
    
    Attributes
    ----------
    auto_completer : `None | SlashCommandParameterAutoCompleter`
        Auto completer if registered.
    
    channel_types : `None | tuple<ChannelType>``
        The accepted channel types.
    
    choice_enum_type : `None | type<Enum>`
        Enum type of `choices` if applicable.
    
    choices : `None | dict<int | float | str | Enum, str>`
        The choices to choose from if applicable. The keys are choice vales meanwhile the values are choice names.
    
    converter : `CoroutineFunctionType`
        The converter to use to convert a value to it's desired type.
    
    default : `object`
        Default value of the parameter.
    
    description : `None | str
        The parameter's description.
    
    max_length : `int`
        The maximum input length allowed for this option.
    
    max_value : `None | int | float`
        The maximal accepted value by the converter.
    
    min_length : `int`
        The minimum input length allowed for this option.
    
    min_value : `None | int | float`
        The minimal accepted value by the converter.
    
    name : `str`
        The parameter's name.
    
    parameter_name : `str`
        The parameter's internal name.
    
    required : `bool`
        Whether the the parameter is required.
    
    type : `int`
        Internal identifier of the converter.
    """
    __slots__ = (
        'auto_completer', 'channel_types', 'choice_enum_type', 'choices', 'converter', 'default', 'description',
        'max_length', 'max_value', 'min_length', 'min_value', 'name', 'required', 'type'
    )
    
    def __new__(
        cls,
        parameter_name,
        converter_type,
        converter,
        name,
        description,
        default,
        required,
        choice_enum_type,
        choices,
        channel_types,
        max_value,
        min_value,
        auto_completer_function,
        max_length,
        min_length,
    ):
        """
        Creates a new ``ParameterConverterSlashCommand`` from the given parameters.
        
        Parameters
        ----------
        parameter_name : `str`
            The parameter's name.
        
        converter_type : `int`
            Internal identifier of the converter.
        
        converter : `CoroutineFunctionType`
            The converter to use to convert a value to it's desired type.
        
        name : `str`
            The parameter's internal name.
        
        description : `None | str`
            The parameter's description.
        
        default : `bool`
            Default value of the parameter.
        
        required : `bool`
            Whether the the parameter is required.
        
        choice_enum_type : `None | type<Enum>`
            Enum type of `choices` if applicable.
        
        choices : `None | dict<int | float | str | Enum, str>`
            The choices to choose from if applicable. The keys are choice vales meanwhile the values are choice names.
        
        channel_types : `None | tuple<ChannelType>`
            The accepted channel types.
        
        max_value : `None | int | float`
            The maximal accepted value by the converter.
        
        min_value : `None | int | float`
            The minimal accepted value by the converter.
        
        auto_completer_function : `None | async-callable`
            Auto completer if defined.
        
        max_length : `int`
            The maximum input length allowed for this option.
        
        min_length : `int`
            The minimum input length allowed for this option.
        """
        self = object.__new__(cls)
        
        self.parameter_name = parameter_name
        self.auto_completer = None
        self.choice_enum_type = choice_enum_type
        self.choices = choices
        self.converter = converter
        self.default = default
        self.description = description
        self.name = name
        self.required = required
        self.type = converter_type
        self.channel_types = channel_types
        self.max_value = max_value
        self.min_value = min_value
        self.max_length = max_length
        self.min_length = min_length
        
        if (auto_completer_function is not None):
            auto_completer = SlashCommandParameterAutoCompleter(
                auto_completer_function, [parameter_name], APPLICATION_COMMAND_FUNCTION_DEEPNESS, None
            )
            self.register_auto_completer(auto_completer)
        
        return self
    
    
    @copy_docs(ParameterConverterBase.__call__)
    async def __call__(self, client, interaction_event, value):
        choices = self.choices
        
        if (value is None):
            if not self.required:
                return self.default
        
        else:
            converted_value = await self.converter(client, interaction_event, value)
            if (converted_value is not None):
                if (choices is None):
                    return converted_value
                
                choice_enum_type = self.choice_enum_type
                if (choice_enum_type is None):
                    if (converted_value in choices):
                        return converted_value
                
                else:
                    try:
                        converted_value = choice_enum_type(converted_value)
                    except ValueError:
                        pass
                    else:
                        return converted_value
        
        
        raise SlashCommandParameterConversionError(
            self.name,
            value,
            ANNOTATION_TYPE_TO_REPRESENTATION.get(self.type, '???'),
            None if choices is None else [*choices.keys()],
        )
    
    @copy_docs(ParameterConverterBase._put_repr_parts_into)
    def _put_repr_parts_into(self, repr_parts):
        # name
        name = self.name
        if (self.parameter_name != name):
            repr_parts.append(', name = ')
            repr_parts.append(repr(name))
        
        # type
        repr_parts.append(', type = ')
        repr_parts.append(ANNOTATION_TYPE_TO_STR_ANNOTATION[self.type])
        
        # description
        repr_parts.append(', description = ')
        repr_parts.append(repr(self.description))
        
        # required
        repr_parts.append(', required = ')
        repr_parts.append(repr(self.required))
        
        if not self.required:
            repr_parts.append(', default = ')
            repr_parts.append(repr(self.default))
        
        # choices
        choices = self.choices
        if (choices is not None):
            repr_parts.append(', choices = ')
            repr_parts.append(repr(choices))
        
        # auto_completer
        auto_completer = self.auto_completer
        if (auto_completer is not None):
            repr_parts.append(', auto_completer = ')
            repr_parts.append(repr(auto_completer))
        
        # channel_types
        channel_types = self.channel_types
        if (channel_types is not None):
            repr_parts.append(', channel_types = ')
            repr_parts.append(repr(channel_types))
        
        # min_value
        min_value = self.min_value
        if (min_value is not None):
            repr_parts.append(', min_value = ')
            repr_parts.append(repr(min_value))
        
        # max_value
        max_value = self.max_value
        if (max_value is not None):
            repr_parts.append(', max_value = ')
            repr_parts.append(repr(max_value))
        
        # min_length
        min_length = self.min_length
        if (min_length != 0):
            repr_parts.append(', min_length = ')
            repr_parts.append(repr(min_length))
        
        # max_length
        max_length = self.max_length
        if (max_length != 0):
            repr_parts.append(', max_length = ')
            repr_parts.append(repr(max_length))
    
    
    @copy_docs(ParameterConverterBase.as_option)
    def as_option(self):
        choices = self.choices
        if choices is None:
            option_choices = None
        else:
            is_enum = (self.choice_enum_type is not None)
            
            option_choices = [
                ApplicationCommandOptionChoice(name, str(value.value if is_enum else value))
                for value, name in choices.items()
            ]
        
        option_type = ANNOTATION_TYPE_TO_OPTION_TYPE[self.type]
        
        return ApplicationCommandOption(
            self.name,
            self.description,
            option_type,
            autocomplete = (self.auto_completer is not None),
            channel_types = self.channel_types,
            choices = option_choices,
            required = self.required,
            min_value = self.min_value,
            max_value = self.max_value,
            min_length = self.min_length,
            max_length = self.max_length,
        )
    
    
    def can_be_auto_completed(self):
        """
        Returns whether the parameter can be auto completed.
        
        Returns
        -------
        can_be_auto_completed : `bool`
            Whether the parameter can be auto completed.
        """
        if (self.type not in ANNOTATION_AUTO_COMPLETE_AVAILABILITY):
            return False
        
        if (self.choices is not None):
            return False
        
        return True
    
    
    def is_auto_completed(self):
        """
        Returns whether the parameter is already auto completed.
        
        Returns
        -------
        is_auto_completed : `bool`
        """
        return (self.auto_completer is not None)
    
    
    def register_auto_completer(self, auto_completer):
        """
        Registers an auto completer to the slash command parameter converter.
        
        Parameters
        ----------
        auto_completer : ``SlashCommandParameterAutoCompleter``
            The auto completer to register.
        
        Returns
        -------
        resolved : `int`
            Whether the parameter was resolved.
        
        Raises
        ------
        RuntimeError
            If the parameter cannot be auto completed.
        """
        if (self.type not in ANNOTATION_AUTO_COMPLETE_AVAILABILITY):
            raise RuntimeError(
                f'Parameter `{self.name}` can not be auto completed. Only string base type parameters '
                f'can be (str, int, expression).'
            )
        
        if (self.choices is not None):
            raise RuntimeError(
                f'Parameter `{self.name}` can not be auto completed. `choices` and `autocomplete` are'
                f'mutually exclusive.'
            )
        
        self_auto_completer = self.auto_completer
        if (self_auto_completer is None) or auto_completer._is_deeper_than(self_auto_completer):
            self.auto_completer = auto_completer
            resolved = 1
        else:
            resolved = 0
        
        return resolved
    
    
    @copy_docs(ParameterConverterBase.bind_parent)
    def bind_parent(self, parent):
        auto_completer = self.auto_completer
        if (auto_completer is not None):
            self.auto_completer = auto_completer._bind_parent(parent)



__all__ = ()

from ...field_parsers import nullable_string_parser_factory
from ...field_putters import nullable_string_putter_factory, url_optional_putter_factory
from ...field_validators import nullable_string_validator_factory, url_optional_validator_factory

from .constants import ICON_URL_LENGTH_MAX, TEXT_LENGTH_MAX

# icon_url

parse_icon_url = nullable_string_parser_factory('icon_url')
put_icon_url = url_optional_putter_factory('icon_url')
# url validator doesnt allow attachment:\\image.png formats
validate_icon_url = nullable_string_validator_factory('icon_url', 0, ICON_URL_LENGTH_MAX)

# icon_proxy_url

parse_icon_proxy_url = nullable_string_parser_factory('proxy_icon_url')
put_icon_proxy_url = url_optional_putter_factory('proxy_icon_url')
validate_icon_proxy_url = url_optional_validator_factory('icon_proxy_url')

# text

parse_text = nullable_string_parser_factory('text')
put_text = nullable_string_putter_factory('text')


def validate_text(text):
    """
    Validates the given embed footer text.
    
    Parameters
    ----------
    text : `None`, `str`, `object`
        Embed footer text.
    
    Returns
    -------
    text : `None`, `str`
    
    Raises
    ------
    TypeError
        - If `text`'s type is incorrect.
    ValueError
        - If `text`'s length is out of the expected range.
    """
    if text is None:
        return None
    
    if not isinstance(text, str):
        text = str(text)
    
    text_length = len(text)
    if text_length == 0:
        return None
    
    if text_length > TEXT_LENGTH_MAX:
        raise ValueError(
            f'`text` length` must be <= {TEXT_LENGTH_MAX}, got {text_length}; text = {text!r}.'
        )
    
    return text



__all__ = ('GuildActivityOverviewTag',)

from scarletio import RichAttributeErrorBaseType

from .fields import parse_emoji, parse_title, put_emoji, put_title, validate_emoji, validate_title


class GuildActivityOverviewTag(RichAttributeErrorBaseType):
    """
    A tag of a guild activity overview.
    
    Attributes
    ----------
    emoji : ``None | Emoji``
        The tag's emoji.
    
    title : `str`
        The tag's emoji
    """
    __slots__ = ('emoji', 'title')
    
    def __new__(cls, *, emoji = ..., title = ...):
        """
        Creates a new guild activity overview tag.
        
        Parameters
        ----------
        emoji : ``None | Emoji``, Optional (Keyword only)
            The tag's emoji.
        
        title : `str`, Optional (Keyword only)
            The tag's emoji
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value in incorrect.
        """
        # emoji
        if emoji is ...:
            emoji = None
        else:
            emoji = validate_emoji(emoji)
        
        # title
        if title is ...:
            title = ''
        else:
            title = validate_title(title)
        
        # Construct
        self = object.__new__(cls)
        self.emoji = emoji
        self.title = title
        return self
    
    
    def __repr__(self):
        """Returns repr(self)."""
        repr_parts = ['<', type(self).__name__]
        
        # emoji
        repr_parts.append(' emoji = ')
        repr_parts.append(repr(self.emoji))
        
        # title
        repr_parts.append(', title = ')
        repr_parts.append(repr(self.title))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __hash__(self):
        """Returns hash(self)."""
        hash_value = 0
        
        # emoji
        emoji = self.emoji
        if (emoji is not None):
            hash_value ^= hash(emoji)
        
        # title
        hash_value ^= hash(self.title)
        
        return hash_value
    
    
    def __eq__(self, other):
        """Returns self == other."""
        if type(self) is not type(other):
            return NotImplemented
        
        # emoji
        if self.emoji is not other.emoji:
            return False
        
        # title
        if self.title != other.title:
            return False
        
        return True
        
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new tag.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Channel data receive from Discord.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self.emoji = parse_emoji(data)
        self.title = parse_title(data)
        return self
    
    
    def to_data(self, *, defaults = False):
        """
        Serializes the guild activity overview tag.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether default field values should be included.
        
        Returns
        -------
        data : `dict<str, object>`
        """
        data = {}
        put_emoji(self.emoji, data, defaults)
        put_title(self.title, data, defaults)
        return data
    
    
    def copy(self):
        """
        Copies the guild activity overview tag.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        new.emoji = self.emoji
        new.title = self.title
        return new
    
    
    def copy_with(self, *, emoji = ..., title = ...):
        """
        Copies the guild activity overview tag with the given fields.
        
        Parameters
        ----------
        emoji : ``None | Emoji``, Optional (Keyword only)
            The tag's emoji.
        
        title : `str`, Optional (Keyword only)
            The tag's emoji
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value in incorrect.
        """
        # emoji
        if emoji is ...:
            emoji = self.emoji
        else:
            emoji = validate_emoji(emoji)
        
        # title
        if title is ...:
            title = self.title
        else:
            title = validate_title(title)
        
        # Construct
        new = object.__new__(type(self))
        new.emoji = emoji
        new.title = title
        return new



__all__ = ('Oauth2User',)

from scarletio import copy_docs

from ...localization.utils import LOCALE_DEFAULT
from ...user import OrinUserBase, PremiumType, UserFlag
from ...user.user.fields import (
    parse_email, parse_email_verified, parse_id, parse_locale, parse_mfa_enabled, parse_premium_type, put_email,
    put_email_verified, put_locale, put_mfa_enabled, put_oauth2_flags_into, put_premium_type,
    validate_email, validate_email_verified, validate_locale, validate_mfa_enabled, validate_premium_type
)

from ..oauth2_access import Oauth2Access


class Oauth2User(OrinUserBase):
    """
    Represents a Discord user with extra personal information. If a ``Oauth2User`` is  created it will NOT overwrite the
    already existing user with the same ID, if exists.
    
    Attributes
    ----------
    access : ``Oauth2Access``
        Source oauth2 access.
    
    avatar_hash : `int`
        The user's avatar's hash in `uint128`.
    
    avatar_type : ``IconType``
        The user's avatar's type.
    
    avatar_decoration : ``None | AvatarDecoration``
        The user's avatar decorations.
    
    banner_color : `None`, ``Color``
        The user's banner color if has any.
    
    banner_hash : `int`
        The user's banner's hash in `uint128`.
    
    banner_type : ``IconType``
        The user's banner's type.
    
    discriminator : `int`
        The user's discriminator. Given to avoid overlapping names.
    
    display_name : `None | str`
        The user's non-unique display name.
    
    email : `None | str`
        The user's email. Defaults to `None`.
    
    email_verified : `bool`
        Whether the email of the user is verified.
    
    flags : ``UserFlag``
        The user's user flags.
    
    id : `int`
        The user's unique identifier number.
    
    locale : ``Locale``
        The preferred locale by the user.
    
    mfa_enabled : `bool`
        Whether the user has two factor authorization enabled on the account.
    
    name : str
        The user's username.
    
    name_plate : ``None | NamePlate``
        The user's name plate.
    
    premium_type : ``PremiumType``
        The Nitro subscription type of the user.
    
    primary_guild_badge : ``None | GuildBadge``
        The user's primary guild's badge.
    """
    __slots__ = ('access', 'email', 'email_verified', 'locale', 'mfa_enabled', 'premium_type')
    
    def __new__(
        cls,
        *,
        avatar = ...,
        avatar_decoration = ...,
        banner = ...,
        banner_color = ...,
        discriminator = ...,
        display_name = ...,
        email = ...,
        email_verified = ...,
        flags = ...,
        locale = ...,
        mfa_enabled = ...,
        name = ...,
        name_plate = ...,
        premium_type = ...,
        primary_guild_badge = ...,
    ):
        """
        Creates a new partial oauth2 user with the given fields.
        
        Parameters
        ----------
        avatar : ``None | str | bytes-like | Icon``, Optional (Keyword only)
            The user's avatar.
        
        avatar_decoration : ``None | AvatarDecoration``, Optional (Keyword only)
            The user's avatar decoration.
        
        banner : ``None | str | bytes-like | Icon``, Optional (Keyword only)
            The user's banner.
        
        banner_color : `None`, ``Color``, `int`, Optional (Keyword only)
            The user's banner color.
        
        discriminator : `str`, `int`, Optional (Keyword only)
            The user's discriminator.
        
        display_name : `None | str`, Optional (Keyword only)
            The user's non-unique display name.
        
        email : `None, `str`, Optional (Keyword only)
            The user's email.
        
        email_verified : `bool`, Optional (Keyword only)
            Whether the email of the user is verified.
        
        flags : `int`, ``UserFlag``, Optional (Keyword only)
            The user's flags.
        
        locale : ``Locale``, `str`, Optional (Keyword only)
            The preferred locale by the user.
        
        mfa_enabled : `bool`, Optional (Keyword only)
            Whether the user has two factor authorization enabled on the account.
        
        name : `str`, Optional (Keyword only)
            The user's name.
        
        name_plate : ``None | NamePlate``, Optional (Keyword only)
            The user's name plate.
        
        premium_type : ``None | int | PremiumType``, Optional (Keyword only)
            The Nitro subscription type of the user.
        
        primary_guild_badge : ``None | GuildBadge``, Optional (Keyword only)
            The user's primary guild's badge.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # email
        if email is ...:
            email = None
        else:
            email = validate_email(email)
        
        # email_verified
        if email_verified is ...:
            email_verified = False
        else:
            email_verified = validate_email_verified(email_verified)
        
        # locale
        if locale is ...:
            locale = LOCALE_DEFAULT
        else:
            locale = validate_locale(locale)
        
        # mfa_enabled
        if mfa_enabled is ...:
            mfa_enabled = False
        else:
            mfa_enabled = validate_mfa_enabled(mfa_enabled)
        
        # premium_type
        if premium_type is ...:
            premium_type = PremiumType.none
        else:
            premium_type = validate_premium_type(premium_type)
        
        # Construct
        self = OrinUserBase.__new__(
            cls,
            avatar = avatar,
            avatar_decoration = avatar_decoration,
            banner = banner,
            banner_color = banner_color,
            discriminator = discriminator,
            display_name = display_name,
            flags = flags,
            name = name,
            name_plate = name_plate,
            primary_guild_badge = primary_guild_badge,
        )
        self.access = Oauth2Access()
        self.email = email
        self.email_verified = email_verified
        self.locale = locale
        self.mfa_enabled = mfa_enabled
        self.premium_type = premium_type
        return self
    
    
    @classmethod
    def from_data(cls, data, access):
        """
        Creates a new ``Oauth2User`` instance.
        
        Parameters
        ----------
        data : `dict<str, object>`
            User data.
        access : ``Oauth2Access``
            Source oauth2 access.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self.access = access
        self.id = parse_id(data)
        self._update_attributes(data)
        return self
    
    
    @copy_docs(OrinUserBase.to_data)
    def to_data(self, *, defaults = False, include_internals = False):
        data = OrinUserBase.to_data(self, defaults = defaults, include_internals = include_internals)
        put_email(self.email, data, defaults)
        put_email_verified(self.email_verified, data, defaults)
        put_locale(self.locale, data, defaults)
        put_mfa_enabled(self.mfa_enabled, data, defaults)
        put_oauth2_flags_into(self.flags, data, defaults)
        put_premium_type(self.premium_type, data, defaults)
        return data
    
    
    @copy_docs(OrinUserBase._update_attributes)
    def _update_attributes(self, data):
        OrinUserBase._update_attributes(self, data)
        
        self.email = parse_email(data)
        self.email_verified = parse_email_verified(data)
        self.locale = parse_locale(data)
        self.mfa_enabled = parse_mfa_enabled(data)
        self.premium_type = parse_premium_type(data)
    
    
    @copy_docs(OrinUserBase._difference_update_attributes)
    def _difference_update_attributes(self, data):
        old_attributes = OrinUserBase._difference_update_attributes(self, data)
        
        email = parse_email(data)
        if self.email != email:
            old_attributes['email'] = self.email
            self.email = email
        
        email_verified = parse_email_verified(data)
        if self.email_verified != email_verified:
            old_attributes['email_verified'] = self.email_verified
            self.email_verified = email_verified
        
        locale = parse_locale(data)
        if self.locale is not locale:
            old_attributes['locale'] = self.locale
            self.locale = locale
        
        mfa_enabled = parse_mfa_enabled(data)
        if self.mfa_enabled != mfa_enabled:
            old_attributes['mfa_enabled'] = self.mfa_enabled
            self.mfa_enabled = mfa_enabled
        
        premium_type = parse_premium_type(data)
        if self.premium_type is not premium_type:
            old_attributes['premium_type'] = self.premium_type
            self.premium_type = premium_type
        
        return old_attributes
    
    
    @copy_docs(OrinUserBase._set_default_attributes)
    def _set_default_attributes(self):
        OrinUserBase._set_default_attributes(self)
        
        self.access = Oauth2Access()
        self.email = None
        self.email_verified = False
        self.locale = LOCALE_DEFAULT
        self.mfa_enabled = False
        self.premium_type = PremiumType.none
    
    
    @copy_docs(OrinUserBase._set_default_attributes)
    def copy(self):
        new = OrinUserBase.copy(self)
        new.access = Oauth2Access()
        new.email = self.email
        new.email_verified = self.email_verified
        new.locale = self.locale
        new.mfa_enabled = self.mfa_enabled
        new.premium_type = self.premium_type
        return new
    
    
    def copy_with(
        self,
        *,
        avatar = ...,
        avatar_decoration = ...,
        banner = ...,
        banner_color = ...,
        discriminator = ...,
        display_name = ...,
        email = ...,
        email_verified = ...,
        flags = ...,
        locale = ...,
        mfa_enabled = ...,
        name = ...,
        name_plate = ...,
        premium_type = ...,
        primary_guild_badge = ...,
    ):
        """
        Creates a new partial oauth2 user with the given fields.
        
        Parameters
        ----------
        avatar : ``None | str | bytes-like | Icon``, Optional (Keyword only)
            The user's avatar.
        
        avatar_decoration : ``None | AvatarDecoration``, Optional (Keyword only)
            The user's avatar decoration.
        
        banner : ``None | str | bytes-like | Icon``, Optional (Keyword only)
            The user's banner.
        
        banner_color : `None`, ``Color``, `int`, Optional (Keyword only)
            The user's banner color.
        
        discriminator : `str`, `int`, Optional (Keyword only)
            The user's discriminator.
        
        display_name : `None | str`, Optional (Keyword only)
            The user's non-unique display name.
        
        email : `None, `str`, Optional (Keyword only)
            The user's email.
        
        email_verified : `bool`, Optional (Keyword only)
            Whether the email of the user is verified.
        
        flags : `int`, ``UserFlag``, Optional (Keyword only)
            The user's flags.
        
        locale : ``Locale``, `str`, Optional (Keyword only)
            The preferred locale by the user.
        
        mfa_enabled : `bool`, Optional (Keyword only)
            Whether the user has two factor authorization enabled on the account.
        
        name : `str`, Optional (Keyword only)
            The user's name.
        
        name_plate : ``None | NamePlate``, Optional (Keyword only)
            The user's name plate.
        
        premium_type : ``None | int | PremiumType``, Optional (Keyword only)
            The Nitro subscription type of the user.
        
        primary_guild_badge : ``None | GuildBadge``, Optional (Keyword only)
            The user's primary guild's badge.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # email
        if email is ...:
            email = self.email
        else:
            email = validate_email(email)
        
        # email_verified
        if email_verified is ...:
            email_verified = self.email_verified
        else:
            email_verified = validate_email_verified(email_verified)
        
        # locale
        if locale is ...:
            locale = self.locale
        else:
            locale = validate_locale(locale)
        
        # mfa_enabled
        if mfa_enabled is ...:
            mfa_enabled = self.mfa_enabled
        else:
            mfa_enabled = validate_mfa_enabled(mfa_enabled)
        
        # premium_type
        if premium_type is ...:
            premium_type = self.premium_type
        else:
            premium_type = validate_premium_type(premium_type)
        
        # Construct
        new = OrinUserBase.copy_with(
            self,
            avatar = avatar,
            avatar_decoration = avatar_decoration,
            banner = banner,
            banner_color = banner_color,
            discriminator = discriminator,
            display_name = display_name,
            flags = flags,
            name = name,
            name_plate = name_plate,
            primary_guild_badge = primary_guild_badge,
        )
        new.access = Oauth2Access()
        new.email = email
        new.email_verified = email_verified
        new.locale = locale
        new.mfa_enabled = mfa_enabled
        new.premium_type = premium_type
        return new
    
    
    @copy_docs(OrinUserBase._get_hash_partial)
    def _get_hash_partial(self):
        hash_value = OrinUserBase._get_hash_partial(self)
        
        # email
        email = self.email
        if (email is not None):
            hash_value ^= hash(email)
        
        # email_verified
        hash_value ^= self.email_verified << 3
        
        # locale
        hash_value ^= hash(self.locale)
        
        # mfa_enabled
        hash_value ^= self.mfa_enabled << 5
        
        # premium_type
        hash_value ^= hash(self.premium_type)
        
        return hash_value
    
    
    def _renew(self, data):
        """
        Renews the oauth2 user's access with the given data.
        
        Parameters
        ----------
        data : `None | dict<str, object>`
            Requested access data.
        """
        self.access._renew(data)
    
    
    # Reflect Oauth2Access
    @property
    def access_token(self):
        """
        Returns the oauth2 user's access's token.
        
        Returns
        -------
        access_token : `str`
        """
        return self.access.access_token
    
    
    @property
    def redirect_url(self):
        """
        Returns the oauth2 user's access's redirect url.
        
        Returns
        -------
        redirect_url : `str`
        """
        return self.access.redirect_url
    
    
    @property
    def refresh_token(self):
        """
        Returns the oauth2 user's access's refresh token.
        
        Returns
        -------
        refresh_token : `str`
        """
        return self.access.refresh_token
    
    
    @property
    def scopes(self):
        """
        Returns the oauth2 user's access's scopes.
        
        Returns
        -------
        scopes : `set<str>`
        """
        return self.access.scopes
    
    
    def has_scope(self, scope):
        """
        Returns whether the oauth2 user's access has the given scope.
        
        Parameters
        ----------
        scope : ``Oauth2Scope``, `str`
            The scope to check out.
        
        Returns
        -------
        has_scope : `bool`
        
        Raises
        ------
        TypeError
            - If `scope` is neither `str`, ``Oauth2Scope``.
        """
        return self.access.has_scope(scope)
    
    
    def iter_scopes(self):
        """
        Iterates over the oauth2 user's scopes.
        
        This method is an iterable generator.
        
        Yields
        ------
        scope : ``Oauth2Scope``
        """
        return (yield from self.access.iter_scopes())



__all__ = ()

from .fields import (
    put_description, put_end, put_name, put_privacy_level, put_schedule, put_start,
    put_status, put_target, validate_description, validate_end, validate_name, validate_privacy_level,
    validate_schedule, validate_start, validate_status, validate_target_location, validate_target_stage,
    validate_target_voice
)


SCHEDULED_EVENT_CREATE_FIELD_CONVERTERS = {
    'description': (validate_description, put_description),
    'end': (validate_end, put_end),
    'location': (validate_target_location, put_target),
    'name': (validate_name, put_name),
    'privacy_level': (validate_privacy_level, put_privacy_level),
    'schedule': (validate_schedule, put_schedule),
    'stage': (validate_target_stage, put_target),
    'start': (validate_start, put_start),
    'voice': (validate_target_voice, put_target),
}

SCHEDULED_EVENT_EDIT_FIELD_CONVERTERS = {
    **SCHEDULED_EVENT_CREATE_FIELD_CONVERTERS,
    'status': (validate_status, put_status),
}


def scheduled_event_occasion_overwrite_get(scheduled_event, timestamp):
    """
    Gets the scheduled event occasion overwrite with the given timestamp.
    
    Parameters
    ----------
    scheduled_event : ``ScheduledEvent``
        The scheduled event to update.
    
    timestamp : ``DateTime``
        Timestamp to search for.
    
    Return
    ------
    occasion_overwrite : ``None | ScheduledEventOccasionOverwrite``
    """
    occasion_overwrites = scheduled_event.occasion_overwrites
    if occasion_overwrites is None:
        return
    
    for occasion_overwrite in occasion_overwrites:
        if occasion_overwrite.timestamp == timestamp:
            return occasion_overwrite


def scheduled_event_occasion_overwrite_add(scheduled_event, occasion_overwrite):
    """
    Adds the given timestamp to ``ScheduledEvent.occasion_overwrites``.
    
    Parameters
    ----------
    scheduled_event : ``ScheduledEvent``
        The scheduled event to update.
    
    occasion_overwrite : ``ScheduledEventOccasionOverwrite``
        Occasion overwrite to add.
    """
    occasion_overwrites = scheduled_event.occasion_overwrites
    if occasion_overwrites is None:
        scheduled_event.occasion_overwrites = (occasion_overwrite, )
        return
    
    timestamp = occasion_overwrite.timestamp
    new_occasion_overwrites = []
    
    iterator = iter(occasion_overwrites)
    while True:
        try:
            added_occasion_overwrite = next(iterator)
        except StopIteration:
            new_occasion_overwrites.append(occasion_overwrite)
            break
        
        if added_occasion_overwrite.timestamp < timestamp:
            new_occasion_overwrites.append(added_occasion_overwrite)
            continue
        
        new_occasion_overwrites.append(occasion_overwrite)
        new_occasion_overwrites.append(added_occasion_overwrite)
        new_occasion_overwrites.extend(iterator)
        break
    
    scheduled_event.occasion_overwrites = tuple(new_occasion_overwrites)
    return


def scheduled_event_occasion_overwrite_remove(scheduled_event, timestamp):
    """
    Removes the given timestamp to ``ScheduledEvent.occasion_overwrites``.
    
    Parameters
    ----------
    scheduled_event : ``ScheduledEvent``
        The scheduled event to update.
    
    timestamp : ``DateTime``
        Timestamp to search for.
    
    Return
    ------
    occasion_overwrite : ``None | ScheduledEventOccasionOverwrite``
    """
    occasion_overwrites = scheduled_event.occasion_overwrites
    if occasion_overwrites is None:
        return None
    
    # Do nothing if already missing.
    for occasion_overwrite in occasion_overwrites:
        if timestamp == occasion_overwrite.timestamp:
            break
    else:
        return None
    
    if len(occasion_overwrites) == 1:
        occasion_overwrites = None
    else:
        occasion_overwrites = (*(
            added_occasion_overwrite for added_occasion_overwrite in occasion_overwrites
            if added_occasion_overwrite.timestamp != timestamp
        ),)
    
    scheduled_event.occasion_overwrites = occasion_overwrites
    return occasion_overwrite



__all__ = ('frame_filter',)

from py_compile import __file__ as PY_COMPILE_FILE_PATH

from ..import_overwrite.source_loader import __file__ as PLUGIN_LOADER_SOURCE_LOADER_FILE_PATH
from ..plugin import __file__ as PLUGIN_LOADER_PLUGIN_FILE_PATH
from ..plugin_loader import __file__ as PLUGIN_LOADER_PLUGIN_LOADER_FILE_PATH


def frame_filter(frame):
    """
    Ignores import frames of plugin loading.
    
    Parameters
    ----------
    frame : ``FrameProxyBase``
        The frame to check.
    
    Returns
    -------
    should_show_frame : `bool`
        Whether the frame should be shown.
    """
    should_show_frame = True
    
    file_name = frame.file_name
    name = frame.name
    line = frame.line
    
    if file_name.startswith('<') and file_name.endswith('>'):
        should_show_frame = False
    
    elif file_name == PLUGIN_LOADER_PLUGIN_FILE_PATH:
        if name == '_load':
            if line == 'loaded = self._load_module()':
                should_show_frame = False
        
        elif name == '_load_module':
            if line == 'spec.loader.exec_module(module)':
                should_show_frame = False
    
    elif file_name == PLUGIN_LOADER_PLUGIN_LOADER_FILE_PATH:
        if name == '_plugin_loader_task':
            if line in (
                'module = await KOKORO.run_in_executor(plugin._load)',
                'await entry_point(module)',
                'entry_point(module)',
            ):
                should_show_frame = False
        
        elif name == '_plugin_unloader_task':
            if line in (
                'await exit_point(module)',
                'exit_point(module)',
            ):
                should_show_frame = False
        
        elif name == '_run_maybe_blocking':
            if line == 'return run_coroutine(coroutine, KOKORO)':
                should_show_frame = False
    
    elif file_name == PLUGIN_LOADER_SOURCE_LOADER_FILE_PATH:
        if name == 'exec_module':
            if line in (
                'exec(code, module.__dict__)',
                'import_plugin(self.name)',
            ):
                should_show_frame = False
    
    elif file_name == PY_COMPILE_FILE_PATH:
        if name == 'compile':
            if line in (
                'code = loader.source_to_code(source_bytes, dfile or file,\n'
                '                             _optimize=optimize)'
            ):
                should_show_frame = False
    
    return should_show_frame



__all__ = ()

from reprlib import repr as short_repr

from scarletio import Compound
from scarletio.web_common import FormData

from ...bases import maybe_snowflake_pair
from ...core import GUILDS, STICKERS
from ...http import DiscordApiClient, VALID_STICKER_IMAGE_MEDIA_TYPES
from ...sticker import Sticker, StickerPack, StickerType
from ...sticker.sticker.fields import (
    put_description, put_name, put_tags, validate_description, validate_name, validate_tags
)
from ...utils import MEDIA_TYPE_TO_EXTENSION, get_image_media_type

from ..functionality_helpers import ForceUpdateCache
from ..request_helpers import get_guild_and_id, get_guild_id, get_sticker_and_id, get_sticker_pack_and_id


STICKER_PACK_CACHE = ForceUpdateCache()


class ClientCompoundStickerEndpoints(Compound):
    
    api : DiscordApiClient
    
    
    async def sticker_get(self, sticker, *, force_update = False):
        """
        Gets an sticker by it's id. If the sticker is already loaded updates it.
        
        This method is a coroutine.
        
        Parameters
        ----------
        sticker : ``Sticker``, `int`
            The sticker, who will be requested.
        force_update : `bool` = `False`, Optional (Keyword only)
            Whether the sticker should be requested even if it supposed to be up to date.
        
        Returns
        -------
        sticker : ``Sticker``
        
        Raises
        ------
        TypeError
            If `sticker` was not given neither as ``Sticker``, neither as `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        TypeError
            If `sticker` was not given as ``Sticker``, nor as `int`.
        """
        sticker, sticker_id = get_sticker_and_id(sticker)
        if (sticker is not None) and (not sticker.partial) and (not force_update):
            return sticker
        
        data = await self.api.sticker_get(sticker_id)
        if (sticker is None):
            sticker = Sticker.from_data(data)
        else:
            sticker._set_attributes(data)
        
        return sticker
    
    
    async def sticker_pack_get(self, sticker_pack, *, force_update = False):
        """
        Gets the sticker packs. If the sticker-packs are already loaded, updates them.
        
        This method is a coroutine.
        
        Parameters
        ----------
        sticker_pack : ``StickerPack``, `int`
            The sticker pack' identifier.
        force_update : `bool` = `False`, Optional (Keyword only)
            Whether the sticker-pack should be requested even if it supposed to be up to date.
        
        Returns
        -------
        sticker_packs : `list` of ``StickerPack``
        
        Raises
        ------
        TypeError
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        sticker_pack, sticker_pack_id = get_sticker_pack_and_id(sticker_pack)
        
        if (sticker_pack is None) or force_update:
            sticker_pack_data = await self.api.sticker_pack_get(sticker_pack_id)
            sticker_pack = StickerPack.from_data(sticker_pack_data, force_update = True)
        
        return sticker_pack
    
    
    async def sticker_pack_get_all(self, *, force_update = False):
        """
        Gets the sticker packs. If the sticker-packs are already loaded, updates them.
        
        This method is a coroutine.
        
        Parameters
        ----------
        force_update : `bool` = `False`, Optional (Keyword only)
            Whether the sticker-packs should be requested even if it supposed to be up to date.
        
        Returns
        -------
        sticker_packs : `list` of ``StickerPack``
        
        Raises
        ------
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        if force_update or (not STICKER_PACK_CACHE.synced):
            data = await self.api.sticker_pack_get_all()
            sticker_pack_datas = data['sticker_packs'] # Discord pls.
            sticker_packs = [
                StickerPack._create_and_update(sticker_pack_data, force_update = force_update)
                for sticker_pack_data in sticker_pack_datas
            ]
            
            STICKER_PACK_CACHE.set(sticker_packs)
        
        else:
            sticker_packs = STICKER_PACK_CACHE.value
        
        return sticker_packs
    
    
    async def sticker_get_guild(self, sticker, *, force_update = False):
        """
        Gets the specified sticker from the respective guild.
        
        This method is a coroutine.
        
        Parameters
        ----------
        sticker : ``Sticker``, `tuple`, (`int`, `int`)
            The sticker to get.
        force_update : `bool` = `False`, Optional (Keyword only)
            Whether the sticker should be requested even if it supposed to be up to date.
        
        Raises
        ------
        TypeError
            If `sticker` is not ``Sticker``, `tuple` (`int`, `int`).
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        if isinstance(sticker, Sticker):
            guild_id = sticker.guild_id
            sticker_id = sticker.id
        else:
            snowflake_pair = maybe_snowflake_pair(sticker)
            if snowflake_pair is None:
                raise TypeError(
                    f'`sticker` can be `{Sticker.__name__}`, `tuple` (`int`, `int`), '
                    f'got {sticker.__class__.__name__}; {sticker!r}.'
                )
            
            guild_id, sticker_id = snowflake_pair
            sticker = STICKERS.get(sticker_id, None)
        
        if (sticker is not None) and (not sticker.partial) and (not force_update):
            return sticker
        
        data = await self.api.sticker_get_guild(guild_id, sticker_id)
        if (sticker is None):
            sticker = Sticker.from_data(data)
        else:
            sticker._set_attributes(data)
        
        return sticker
    
    
    async def sticker_create(self, guild, name, image, tags = None, description = None, *, reason = None):
        """
        Creates a sticker in the guild.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``int | Guild``
            The guild to create the sticker in.
        name : `str`
            The sticker's name.
        tags : `None | str | iterable<str>`
            The tags of the sticker.
        image : `bytes-like`
            The sticker's image in bytes.
        description : `None`, `str` = `None`, Optional
            The sticker's representation.
        reason : `None`, `str` = `None`, Optional (Keyword only)
            Will show up at the respective guild's audit logs.
        
        Returns
        -------
        sticker : ``Sticker``
            The created sticker.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        ValueError
            If `image`s media type is neither `image/png` nor `application/json`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        guild_id = get_guild_id(guild)
        
        description = validate_description(description)
        name = validate_name(name)
        tags = validate_tags(tags)
        
        if not isinstance(image, (bytes, bytearray, memoryview)):
            raise TypeError(
                f'`image` can be `None`, `bytes-like`, got {image.__class__.__name__}; {short_repr(image)}.'
            )
        
        media_type = get_image_media_type(image)
        if media_type not in VALID_STICKER_IMAGE_MEDIA_TYPES:
            raise ValueError(
                f'Invalid `image` type: {media_type}, got {short_repr(image)}.'
            )
        
        extension = MEDIA_TYPE_TO_EXTENSION[media_type]
        
        form_data = FormData()
        form_data.add_field('name', name)
        # If no description is given Discord drops back an unrelated error
        form_data.add_field('description', '' if description is None else description)
        form_data.add_field('tags', '' if tags is None else ', '.join(tags))
        form_data.add_field('file', image, file_name = f'file.{extension}', content_type = media_type)
        
        sticker_data = await self.api.sticker_create(guild_id, form_data, reason)
        
        return Sticker.from_data(sticker_data)
    
    
    async def sticker_edit(self, sticker, *, name = ..., tags = ..., description = ..., reason = None):
        """
        Edits the given guild bound sticker,
        
        This method is a coroutine.
        
        Parameters
        ----------
        sticker : ``Sticker``, `int`
            The respective sticker.
        name : `str`, Optional (keyword only)
            New name of the sticker. It's length can be in range [2:32].
        tags : `str`, ``Emoji``, `iterable` of `str`, Optional (keyword Only)
            The new emoji representation of the sticker. Used as a tag for the sticker.
        description : `None`, `str`, Optional (Keyword only)
            New description for the sticker. It's length can be in range [0:100].
        reason : `None`, `str` = `None`, Optional (Keyword only)
            Will show up at the respective guild's audit logs.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
            - Standard stickers cannot be edited.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        sticker = await self.sticker_get(sticker)
        
        
        if sticker.type is StickerType.standard:
            raise ValueError(
                f'Standard sticker cannot be edited, got {sticker!r}.'
            )
        
        if (name is ...):
            name = sticker.name
        else:
            name = validate_name(name)
        
        if (tags is ...):
            tags = sticker.tags
        else:
            tags = validate_tags(tags)
        
        if (description is ...):
            description = sticker.description
        else:
            description = validate_description(description)
        
        data = {}
        put_description(description, data, True)
        put_name(name, data, True)
        put_tags(tags, data, True)
        
        await self.api.sticker_edit(sticker.guild_id, sticker.id, data, reason)
    
    
    async def sticker_delete(self, sticker, *, reason = None):
        """
        Deletes the sticker.
        
        This method is a coroutine.
        
        Parameters
        ----------
        sticker : ``Sticker``, `int`
            The sticker to delete.
        reason : `None`, `str` = `None`, Optional (Keyword only)
            Will show up at the respective guild's audit logs.
        
        Raises
        ------
        TypeError
            If `sticker` is neither ``Sticker``, nor `int`.
        ValueError
            - Standard stickers cannot be deleted.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        sticker = await self.sticker_get(sticker)
        
        if sticker.type is StickerType.standard:
            raise ValueError(
                f'Standard sticker cannot be deleted, got {sticker!r}.'
            )
        
        await self.api.sticker_delete(sticker.guild_id, sticker.id, reason)
    
    
    async def sticker_get_all_guild(self, guild):
        """
        Syncs the given guild's stickers with the wrapper.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``int | Guild``
            The guild, what's stickers will be synced.
        
        Returns
        -------
        stickers : `list` of ``Sticker``
            The guild's stickers.
        
        Raises
        ------
        TypeError
            If `guild` was not given neither as ``Guild`` nor as `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        guild, guild_id = get_guild_and_id(guild)
        
        sticker_datas = await self.api.sticker_get_all_guild(guild_id)
        
        if guild is None:
            # Do not create a partial guild, because it would have been garbage collected after leaving the function
            # anyways
            guild = GUILDS.get(guild_id, None)
        
        if guild is None:
            stickers = [Sticker.from_data(sticker_data) for sticker_data in sticker_datas]
        else:
            guild._update_stickers(sticker_datas)
            stickers = [*guild.stickers.values()]
        
        return stickers



__all__ = ()

from ...bases import Icon
from ...user import GuildProfileFlag
from ...user.guild_profile.fields import (
    validate_bypasses_verification, validate_flags, validate_nick, validate_pending, validate_timed_out_until
)
from ...user.guild_profile.guild_profile import GUILD_PROFILE_AVATAR
from ...user.voice_state.fields import validate_deaf, validate_mute
from ...utils import datetime_to_timestamp, timestamp_to_datetime

from ..audit_log_entry_change_conversion import AuditLogEntryChangeConversion, AuditLogEntryChangeConversionGroup
from ..audit_log_entry_change_conversion.change_deserializers import change_deserializer_addition_and_removal
from ..audit_log_entry_change_conversion.change_serializers import change_serializer_addition_and_removal
from ..audit_log_entry_change_conversion.value_mergers import value_merger_sorted_array
from ..audit_log_role import AuditLogRole
from ..conversion_helpers.converters import value_deserializer_description, value_serializer_description


# ---- avatar ----

AVATAR_CONVERSION = AuditLogEntryChangeConversion(
    ('avatar_hash',),
    'avatar',
    value_deserializer = Icon.from_base_16_hash,
    value_serializer = Icon.as_base_16_hash.fget,
    value_validator = GUILD_PROFILE_AVATAR.validate_icon,
)


# ---- bypasses_verification ----

BYPASSES_VERIFICATION_CONVERSION = AuditLogEntryChangeConversion(
    ('bypasses_verification',),
    'bypasses_verification',
    value_validator = validate_bypasses_verification,
)


@BYPASSES_VERIFICATION_CONVERSION.set_value_deserializer
def bypasses_verification_value_deserializer(value):
    if value is None:
        value = False

    return value


# ---- deaf ----

DEAF_CONVERSION = AuditLogEntryChangeConversion(
    ('deaf',),
    'deaf',
    value_validator = validate_deaf,
)


@DEAF_CONVERSION.set_value_deserializer
def deaf_value_deserializer(value):
    if value is None:
        value = False

    return value


# ---- flags ----

FLAGS_CONVERSION = AuditLogEntryChangeConversion(
    ('flags',),
    'flags',
    value_validator = validate_flags,
)


@FLAGS_CONVERSION.set_value_deserializer
def flags_value_deserializer(value):
    if value is None:
        value = GuildProfileFlag()
    else:
        value = GuildProfileFlag(value)
    
    return value


@FLAGS_CONVERSION.set_value_serializer
def flags_value_serializer(value):
    return int(value)


# ---- mute ----

MUTE_CONVERSION = AuditLogEntryChangeConversion(
    ('mute',),
    'mute',
    value_validator = validate_mute,
)


@MUTE_CONVERSION.set_value_deserializer
def mute_value_deserializer(value):
    if value is None:
        value = False

    return value


# ---- nick ----

NICK_CONVERSION = AuditLogEntryChangeConversion(
    ('nick',),
    'nick',
    value_deserializer = value_deserializer_description,
    value_serializer = value_serializer_description,
    value_validator = validate_nick,
)


# ---- pending ----

PENDING_CONVERSION = AuditLogEntryChangeConversion(
    ('pending',),
    'pending',
    value_validator = validate_pending,
)


@PENDING_CONVERSION.set_value_deserializer
def pending_value_deserializer(value):
    if value is None:
        value = False

    return value


# ---- roles ----

ROLES_CONVERSION = AuditLogEntryChangeConversion(
    ('$remove', '$add'),
    'roles',
    change_deserializer = change_deserializer_addition_and_removal,
    change_serializer = change_serializer_addition_and_removal,
    value_merger = value_merger_sorted_array,
)


@ROLES_CONVERSION.set_value_deserializer
def roles_value_deserializer(value):
    if value is None:
        pass
    elif (not value):
        value = None
    else:
        value = (*sorted(AuditLogRole.from_data(role_data) for role_data in value),)
    return value


@ROLES_CONVERSION.set_value_serializer
def roles_value_serializer(value):
    if value is None:
        value = []
    else:
        value = [role.to_data(defaults = True) for role in value]
    return value


@ROLES_CONVERSION.set_value_validator
def roles_value_validator(value):
    if value is None:
        return None
    
    if (getattr(type(value), '__iter__', None) is None):
        raise TypeError(
            f'`roles` can be `None`, `iterable` of `{AuditLogRole.__name__}`, got {type(value).__name__}, {value!r}.'
        )
    
    roles = None
    for role in value:
        if not isinstance(role, AuditLogRole):
            raise TypeError(
                f'`roles` can contain `{AuditLogRole.__name__}` elements, got {type(role).__name__}; {role!r}; '
                f'roles = {roles!r}.'
            )
        
        if roles is None:
            roles = []
        roles.append(role)
    
    if (roles is not None):
        roles.sort()
        roles = (*roles,)
    
    return roles


# ---- timed_out_until ----

TIMED_OUT_UNTIL_CONVERSION = AuditLogEntryChangeConversion(
    ('communication_disabled_until',),
    'timed_out_until',
    value_validator = validate_timed_out_until,
)


@TIMED_OUT_UNTIL_CONVERSION.set_value_deserializer
def timed_out_until_converter_get(value):
    if (value is not None):
        value = timestamp_to_datetime(value)
    return value


@TIMED_OUT_UNTIL_CONVERSION.set_value_serializer
def timed_out_until_converter_put(value):
    if (value is not None):
        value = datetime_to_timestamp(value)
    return value


# ---- Construct ----

USER_CONVERSIONS = AuditLogEntryChangeConversionGroup(
    AVATAR_CONVERSION,
    BYPASSES_VERIFICATION_CONVERSION,
    DEAF_CONVERSION,
    FLAGS_CONVERSION,
    MUTE_CONVERSION,
    NICK_CONVERSION,
    PENDING_CONVERSION,
    ROLES_CONVERSION,
    TIMED_OUT_UNTIL_CONVERSION,
)



__all__ = ()

from os import listdir as list_directory, mkdir as create_directory
from os.path import (
    abspath as absolute_path, basename as base_name, dirname as get_parent_directory_path, exists, isdir as is_directory
)

from .layouts import DEFAULT_LAYOUT, LAYOUT_DESCRIPTIONS


def _validate_bot(bots_value):
    """
    Validates the given `bot` value. Either `directory_path` or `error_message` is always returned as non-`None`.
    
    Parameters
    ----------
    bots_value : `tuple` of `str`
        The inputted bots.
    
    Returns
    -------
    bots : `None | list<str>`
        The validated bots.
    error_message : `None`, `str`
        Error message.
    """
    if not bots_value:
        return None, f'Please define at least one bot to create.\n'
    
    for value in bots_value:
        if (not value) or (not value.isidentifier()):
            return None, f'Bot name: {value!r} is not a valid identifier.\n'
    
    return sorted({*bots_value}), None


def _validate_name(name_value):
    """
    Validates the given `name` value. Either `directory_path` or `error_message` is always returned as non-`None`.
    
    Parameters
    ----------
    name_value : `str`
        The passed value by the user.
    
    Returns
    -------
    directory_path : `None`, `str`
        Directory path of the project. Might need to be created recursively.
    error_message : `None`, `str`
        Error message.
    """
    directory_path = absolute_path(name_value)
    if is_directory(directory_path):
        if list_directory(directory_path):
            return None, f'Path already exists and not empty:\n{directory_path}\n'
    
    elif exists(directory_path):
        return None, f'Path already exists and its not a directory:\n{directory_path}\n'
    
    else:
        parent_directory_path = get_parent_directory_path(directory_path)
        while True:
            if is_directory(parent_directory_path):
                break
            
            elif exists(parent_directory_path):
                return None, f'Parent already exists, but its not a directory:\n{parent_directory_path}\n'
            
            parent_directory_path = get_parent_directory_path(parent_directory_path)
            continue
    
    return directory_path, None


def _validate_project_name(project_name_value, name_value):
    """
    Validates the given `project_name` value.
    
    Either `project_name` or `error_message` is always returned as non-`None`.
    
    Parameters
    ----------
    project_name_value : `None`, `str`
        The project's name value explicitly defined by the user if any.
    name_value : `str`
        The project's name (used for location).
    
    Returns
    -------
    project_name : `None`, `str`
        The project's name.
    error_message : `None`, `str`
        Error message.
    """
    if (project_name_value is not None) and project_name_value:
        project_name = project_name_value
    else:
        project_name = base_name(name_value)
    
    if (not project_name) or (not project_name.isidentifier()):
        return (
            None,
            (
                f'A project name\'s must be a valid identifier, got {project_name!r}.\n'
                f'Note that if project name is not explicitly defined it is detected from the name parameter.\n'
            )
        )
    
    return project_name, None


def _build_identifying_layout_failed_error_message(layout_value):
    """
    Builds error message for ``_validate_layout`` when identifying the layout failed.
    
    Parameters
    ----------
    layout_value : `str`
        The passed layout value.
    
    Returns
    -------
    error_message : `str`
        Error message.
    """
    error_message_parts = []
    
    error_message_parts.append('Could not identify layout: ')
    error_message_parts.append(repr(layout_value))
    error_message_parts.append('\nPlease use on of the following:\n')
    
    for layout in sorted(LAYOUT_DESCRIPTIONS.keys()):
        error_message_parts.append('- ')
        error_message_parts.append(layout)
        error_message_parts.append('\n')
    
    return ''.join(error_message_parts)


def _validate_layout(layout_value):
    """
    Validates the given `layout` value.
    
    Parameters
    ----------
    layout : `None`, `str`
        The passed layout value.
    
    Returns
    -------
    layout : `str`
        The layout's name.
    error_message : `None`, `str`
        Error message.
    """
    if (layout_value is None) or (not layout_value):
        layout = DEFAULT_LAYOUT
    else:
        layout = layout_value.casefold()
        if layout not in LAYOUT_DESCRIPTIONS.keys():
            return None, _build_identifying_layout_failed_error_message(layout_value)
    
    return layout, None


def create_directory_recursive(directory_path):
    """
    Creates the directory's path recursively.
    
    Parameters
    ----------
    directory_path : `str`
        The path to create.
    """
    to_create = None
    
    while True:
        if exists(directory_path):
            break
        
        if (to_create is None):
            to_create = []
        to_create.append(directory_path)
        
        directory_path = get_parent_directory_path(directory_path)
        continue
    
    if (to_create is not None):
        for path in reversed(to_create):
            create_directory(path)



__all__ = ('CONVERSION_INTERACTION_EVENT',)

from ....discord.builder.conversion import Conversion
from ....discord.builder.constants import CONVERSION_KIND_FIELD
from ....discord.interaction import InteractionEvent


class CONVERSION_INTERACTION_EVENT(Conversion):
    # Generic
    
    name = 'interaction_event'
    name_aliases = ('event', )
    expected_types_messages = f'`None`, `{InteractionEvent.__name__}`'
    kind = CONVERSION_KIND_FIELD
    output_conversion = None
    
    # Setting
    
    set_merger = None
    set_type = InteractionEvent
    set_type_processor = None
    set_identifier = None
    set_listing_identifier = None
    
    
    def set_validator(value):
        # None
        if value is None:
            yield None
            return
        
        # InteractionEvent
        if isinstance(value, InteractionEvent):
            yield value
            return
        
        # No other cases
        return
    
    
    # Reading
    
    get_default = None
    get_processor = None
    
    
    # Serialization
    
    serializer_key = None
    serializer_optional = None
    serializer_required = None
    
    # Sorting
    
    sort_priority = -1



__all__ = ()

from ...field_parsers import (
    bool_parser_factory, entity_id_parser_factory, int_parser_factory, preinstanced_parser_factory
)
from ...field_putters import (
    entity_id_putter_factory, force_bool_putter_factory, int_putter_factory, preinstanced_putter_factory
)
from ...field_validators import (
    bool_validator_factory, entity_id_validator_factory, int_conditional_validator_factory,
    preinstanced_validator_factory
)

from ..client_platform_configuration import ClientPlatformConfiguration

from .preinstanced import OrientationLockState, PlatformType


# age_gated

parse_age_gated = bool_parser_factory('requires_age_gate', False)
put_age_gated = force_bool_putter_factory('requires_age_gate')
validate_age_gated = bool_validator_factory('age_gated', False)


# client_platform_configurations

def parse_client_platform_configurations(data):
    """
    Parses out client platform configurations from the given data.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Embedded activity configuration data.
    
    Returns
    -------
    client_platform_configurations : `None | dict<PlatformType, ClientPlatformConfiguration>`
    """
    configuration_datas = data.get('client_platform_config', None)
    if (configuration_datas is None) or (not configuration_datas):
        return None
    
    return {
        PlatformType(platform_value): ClientPlatformConfiguration.from_data(configuration_data)
        for platform_value, configuration_data in configuration_datas.items()
    }
        

def put_client_platform_configurations(client_platform_configurations, data, defaults):
    """
    Puts the activity timestamps start into the given data.
    
    Parameters
    ----------
    client_platform_configurations : `None | dict<PlatformType, ClientPlatformConfiguration>`
        Client configurations by platform.
    data : `dict<str, object>`
        Json serializable dictionary.
    defaults : `bool`
        Whether default values should be included as well.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    configuration_datas = {}
    supported_platforms = []
    
    if (client_platform_configurations is not None):
        for platform, client_platform_configuration in client_platform_configurations.items():
            configuration_datas[platform.value] = client_platform_configuration.to_data(defaults = defaults)
            supported_platforms.append(platform.value)
    
    supported_platforms.sort()
    
    data['client_platform_config'] = configuration_datas
    data['supported_platforms'] = supported_platforms
    
    return data


def validate_client_platform_configurations(client_platform_configurations):
    """
    Validates the given client platform configurations.
    
    Parameters
    ----------
    client_platform_configurations : `None | dict<PlatformType | int, ClientPlatformConfiguration>`
        Client configurations by platform.
    
    Returns
    -------
    client_platform_configurations : `None | dict<PlatformType, ClientPlatformConfiguration>`
    
    Raises
    ------
    TypeError
        - If a parameter's type is incorrect.
    ValueError
        - If a parameter's value is incorrect.
    """
    if client_platform_configurations is None:
        return None
    
    if not isinstance(client_platform_configurations, dict):
        raise TypeError(
            f'`client_platform_configurations` can be `None | dict<{PlatformType.__name__} | '
            f'{PlatformType.VALUE_TYPE.__name__}, {ClientPlatformConfiguration.__name__}`, got '
            f'{type(client_platform_configurations).__name__}; {client_platform_configurations!r}.'
        )
    
    validated = None
    
    for key, value in client_platform_configurations.items():
        if isinstance(key, PlatformType):
            pass
        elif isinstance(key, PlatformType.VALUE_TYPE):
            key = PlatformType(key)
        else:
            raise TypeError(
                f'`client_platform_configurations` keys can be `{PlatformType.__name__} | '
                f'{PlatformType.VALUE_TYPE.__name__}`, got {type(key).__name__}; {key!r}.'
            )
        
        if not isinstance(value, ClientPlatformConfiguration):
            raise TypeError(
                f'`client_platform_configurations` values can be `{ClientPlatformConfiguration.__name__}`, got '
                f'{type(value).__name__}; {value!r}.'
            )
        
        if validated is None:
            validated = {}
        
        validated[key] = value
    
    return validated


# content_security_policy_exceptions_exist

parse_content_security_policy_exceptions_exist = bool_parser_factory('has_csp_exception', False)
put_content_security_policy_exceptions_exist = force_bool_putter_factory('has_csp_exception')
validate_content_security_policy_exceptions_exist = bool_validator_factory(
    'content_security_policy_exceptions_exist', False
)


# default_orientation_lock_state

parse_default_orientation_lock_state = preinstanced_parser_factory(
    'default_orientation_lock_state', OrientationLockState, OrientationLockState.none
)
put_default_orientation_lock_state = preinstanced_putter_factory('default_orientation_lock_state')
validate_default_orientation_lock_state = preinstanced_validator_factory(
    'default_orientation_lock_state', OrientationLockState
)


# default_tablet_orientation_lock_state

parse_default_tablet_orientation_lock_state = preinstanced_parser_factory(
    'tablet_default_orientation_lock_state', OrientationLockState, OrientationLockState.none
)
put_default_tablet_orientation_lock_state = preinstanced_putter_factory('tablet_default_orientation_lock_state')
validate_default_tablet_orientation_lock_state = preinstanced_validator_factory(
    'default_tablet_orientation_lock_state', OrientationLockState
)


# position

parse_position = int_parser_factory('shelf_rank', 0)
put_position = int_putter_factory('shelf_rank')
validate_position = int_conditional_validator_factory(
    'position',
    0,
    (lambda position : position >= 0),
    '>= 0',
)

# preview_video_asset_id

parse_preview_video_asset_id = entity_id_parser_factory('activity_preview_video_asset_id')
put_preview_video_asset_id = entity_id_putter_factory('activity_preview_video_asset_id')
validate_preview_video_asset_id = entity_id_validator_factory('preview_video_asset_id')




__all__ = ('ClientPlatformConfiguration',)

from scarletio import RichAttributeErrorBaseType

from ...utils import DATETIME_FORMAT_CODE

from .fields import (
    parse_label_type, parse_labelled_until, parse_release_phase, put_label_type, put_labelled_until,
    put_release_phase, validate_label_type, validate_labelled_until, validate_release_phase
)
from .preinstanced import LabelType, ReleasePhase


class ClientPlatformConfiguration(RichAttributeErrorBaseType):
    """
    Represents an embedded activity's configuration on a client's platform.
    
    Attributes
    ----------
    label_type : ``LabelType``
        The label's type on the client's platform.
    labelled_until : `None | DateTime`
        Till when is the label shown.
    release_phase : ``ReleasePhase``
        What release phase is the embedded activity is on the client's platform.
    """
    __slots__ = ('label_type', 'labelled_until', 'release_phase', )
    
    def __new__(cls, *, label_type = ..., labelled_until = ..., release_phase = ...):
        """
        Creates a new client platform configuration.
        
        Parameters
        ----------
        label_type : `LabelType | int`, Optional (Keyword only)
            The label's type on the client's platform.
        labelled_until : `None | DateTime`, Optional (Keyword only)
            Till when is the label shown.
        release_phase : `ReleasePhase | str`, Optional (Keyword only)
            What release phase is the embedded activity is on the client's platform.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's type is incorrect.
        """
        # label_type
        if label_type is ...:
            label_type = LabelType.none
        else:
            label_type = validate_label_type(label_type)
        
        # labelled_until
        if labelled_until is ...:
            labelled_until = None
        else:
            labelled_until = validate_labelled_until(labelled_until)
        
        # release_phase
        if release_phase is ...:
            release_phase = ReleasePhase.global_launch
        else:
            release_phase = validate_release_phase(release_phase)
        
        # Construct
        self = object.__new__(cls)
        self.label_type = label_type
        self.labelled_until = labelled_until
        self.release_phase = release_phase
        return self
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a client platform configuration from the given data.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Client platform configuration data.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self.label_type = parse_label_type(data)
        self.labelled_until = parse_labelled_until(data)
        self.release_phase = parse_release_phase(data)
        return self
    
    
    def to_data(self, *, defaults = False):
        """
        Serializes the client platform configuration.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether fields of their default value should be included as well.
        
        Returns
        -------
        data : `dict<str, object>`
        """
        data = {}
        put_label_type(self.label_type, data, defaults)
        put_labelled_until(self.labelled_until, data, defaults)
        put_release_phase(self.release_phase, data, defaults)
        return data
    
    
    def __repr__(self):
        """Returns the client platform configuration's representation."""
        repr_parts = [
            '<',
            self.__class__.__name__,
        ]
        
        label_type = self.label_type
        repr_parts.append(' label_type = ')
        repr_parts.append(label_type.name)
        repr_parts.append(' ~ ')
        repr_parts.append(repr(label_type.value))
        
        
        labelled_until = self.labelled_until
        repr_parts.append(', labelled_until = ')
        if labelled_until is None:
            repr_parts.append('None')
        else:
            repr_parts.append(format(labelled_until, DATETIME_FORMAT_CODE))
        
        release_phase = self.release_phase
        repr_parts.append(', release_phase = ')
        repr_parts.append(release_phase.name)
        repr_parts.append(' ~ ')
        repr_parts.append(repr(release_phase.value))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __eq__(self, other):
        """Returns whether the two client platform configurations are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        if self.label_type is not other.label_type:
            return False
        
        if self.labelled_until != other.labelled_until:
            return False
        
        if self.release_phase is not other.release_phase:
            return False
        
        return True
    

    def __hash__(self):
        """Returns the client platform configuration's hash value."""
        hash_value = 0
        
        hash_value ^= hash(self.label_type)
        
        labelled_until = self.labelled_until
        if (labelled_until is not None):
            hash_value ^= hash(labelled_until)
        
        hash_value ^= hash(self.release_phase) << 4
        
        return hash_value
    
    
    def copy(self):
        """
        Copies the client platform configuration.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        new.label_type = self.label_type
        new.labelled_until = self.labelled_until
        new.release_phase = self.release_phase
        return new
    
    
    def copy_with(self, *, label_type = ..., labelled_until = ..., release_phase = ...):
        """
        Copies the client platform configuration with the given fields.
        
        Parameters
        ----------
        label_type : `LabelType | int`, Optional (Keyword only)
            The label's type on the client's platform.
        labelled_until : `None | DateTime`, Optional (Keyword only)
            Till when is the label shown.
        release_phase : `ReleasePhase | str`, Optional (Keyword only)
            What release phase is the embedded activity is on the client's platform.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's type is incorrect.
        """
        # label_type
        if label_type is ...:
            label_type = self.label_type
        else:
            label_type = validate_label_type(label_type)
        
        # labelled_until
        if labelled_until is ...:
            labelled_until = self.labelled_until
        else:
            labelled_until = validate_labelled_until(labelled_until)
        
        # release_phase
        if release_phase is ...:
            release_phase = self.release_phase
        else:
            release_phase = validate_release_phase(release_phase)
        
        # Construct
        new = object.__new__(type(self))
        new.label_type = label_type
        new.labelled_until = labelled_until
        new.release_phase = release_phase
        return new



__all__ = ('ComponentType',)

from scarletio import export

from ...bases import Preinstance as P, PreinstancedBase
from ...resolved.resolver.resolvers import (
    RESOLVER_ATTACHMENT, RESOLVER_CHANNEL, RESOLVER_MENTIONABLE, RESOLVER_ROLE, RESOLVER_STRING, RESOLVER_USER
)

from ..component_metadata import (
    ComponentMetadataAttachmentInput, ComponentMetadataAttachmentMedia, ComponentMetadataBase, ComponentMetadataButton,
    ComponentMetadataChannelSelect, ComponentMetadataContainer, ComponentMetadataLabel, ComponentMetadataMediaGallery,
    ComponentMetadataMentionableSelect, ComponentMetadataRoleSelect, ComponentMetadataRow, ComponentMetadataSection,
    ComponentMetadataSeparator, ComponentMetadataStringSelect, ComponentMetadataTextDisplay, ComponentMetadataTextInput,
    ComponentMetadataThumbnailMedia, ComponentMetadataUserSelect
)
from ..interaction_component_metadata import (
    InteractionComponentMetadataAttachmentInput, InteractionComponentMetadataBase, InteractionComponentMetadataButton,
    InteractionComponentMetadataChannelSelect, InteractionComponentMetadataContainer, InteractionComponentMetadataLabel,
    InteractionComponentMetadataMentionableSelect, InteractionComponentMetadataRoleSelect,
    InteractionComponentMetadataRow, InteractionComponentMetadataSection, InteractionComponentMetadataStringSelect,
    InteractionComponentMetadataTextInput, InteractionComponentMetadataUserSelect
)

from .flags import ComponentTypeLayoutFlag


COMPONENT_TYPE_LAYOUT_FLAGS_ALL = ComponentTypeLayoutFlag().update_by_keys(
    allowed_in_message = True,
    allowed_in_form = True,
    top_level = True,
    nestable_into_row = True,
    nestable_into_container = True,
    nestable_into_section = True,
    section_thumbnail = True,
    version_1 = True,
    version_2 = True,
    nestable_into_label = True,
    holds_value_single = True,
    holds_value_multiple = True,
)


@export
class ComponentType(PreinstancedBase, value_type = int):
    """
    Represents a component's type.
    
    Attributes
    ----------
    layout_flags : ``ComponentTypeLayoutFlag``
        Flags about the component's layout information.
    
    interaction_metadata_type : ``type<InteractionComponentMetadataBase>``
        Interaction component metadata type.
    
    iter_resolve : `None | GeneratorFunctionType`
        Iterative resolver function for resolving the values held by interaction components of this type.
    
    metadata_type : ``type<ComponentMetadataBase>``
        Metadata type.
    
    resolve : `None | FunctionType`
        Resolver function for resolving the values held by interaction components of this type.
    
    resolver : ``None | Resolver``
        The general resolver holding the resolve function for resolving the values held by interaction components of
        this type.
    
    name : `str`
        The name of the component type.
    
    value : `int`
        The identifier value the component type.
    
    
    Type Attributes
    ---------------
    Every predefined component type can be accessed as type attribute as well:
    
    +-----------------------+-------------------------------+-------+
    | Type attribute name   | Name                          | Value |
    +=======================+===============================+=======+
    | none                  | none                          | 0     |
    +-----------------------+-------------------------------+-------+
    | row                   | row                           | 1     |
    +-----------------------+-------------------------------+-------+
    | button                | button                        | 2     |
    +-----------------------+-------------------------------+-------+
    | string_select         | string select                 | 3     |
    +-----------------------+-------------------------------+-------+
    | text_input            | text input                    | 4     |
    +-----------------------+-------------------------------+-------+
    | user_select           | user select                   | 5     |
    +-----------------------+-------------------------------+-------+
    | role_select           | role select                   | 6     |
    +-----------------------+-------------------------------+-------+
    | mentionable_select    | mentionable select            | 7     |
    +-----------------------+-------------------------------+-------+
    | channel_select        | channel select                | 8     |
    +-----------------------+-------------------------------+-------+
    | section               | section                       | 9     |
    +-----------------------+-------------------------------+-------+
    | text_display          | text display                  | 10    |
    +-----------------------+-------------------------------+-------+
    | thumbnail_media       | thumbnail media               | 11    |
    +-----------------------+-------------------------------+-------+
    | media_gallery         | media gallery                 | 12    |
    +-----------------------+-------------------------------+-------+
    | attachment_media      | attachment media              | 13    |
    +-----------------------+-------------------------------+-------+
    | separator             | separator                     | 14    |
    +-----------------------+-------------------------------+-------+
    | ???                   | ???                           | 15    |
    +-----------------------+-------------------------------+-------+
    | ???                   | activity content inventory    | 16    |
    +-----------------------+-------------------------------+-------+
    | container             | container                     | 17    |
    +-----------------------+-------------------------------+-------+
    | label                 | label                         | 18    |
    +-----------------------+-------------------------------+-------+
    | attachment_input      | attachment input              | 19    |
    +-----------------------+-------------------------------+-------+
    """
    __slots__ = ('layout_flags', 'interaction_metadata_type', 'iter_resolve', 'metadata_type', 'resolve', 'resolver')
    
    
    def __new__(
        cls,
        value,
        name = None,
        *,
        interaction_metadata_type = ...,
        layout_flags = ...,
        metadata_type = ...,
        resolver = ...,
    ):
        """
        Creates a new component type.
        
        Parameters
        ----------
        value : `int`
            The Discord side identifier value of the channel type.
        
        name : `None | str` = `None`, Optional
            The default name of the channel type.
        
        interaction_metadata_type : ``type<InteractionComponentMetadataBase>``, Optional (Keyword only)
            The component type's respective interaction metadata type.
        
        layout_flags : ``ComponentTypeLayoutFlag``, Optional (Keyword only)
            Flags about the component's layout information.
        
        metadata_type : ``type<ComponentMetadataBase>``, Optional (Keyword only)
            The component type's respective metadata type.
        
        resolver : ``Resolver``, Optional (Keyword only)
            The general resolver holding the resolve function for resolving the values held by interaction components
            of this type.
        """
        if interaction_metadata_type is ...:
            interaction_metadata_type = InteractionComponentMetadataBase
        
        if layout_flags is ...:
            layout_flags = COMPONENT_TYPE_LAYOUT_FLAGS_ALL
        
        if metadata_type is ...:
            metadata_type = ComponentMetadataBase
        
        if resolver is ...:
            resolver = None
        
        while True:
            if (resolver is not None):
                if layout_flags.holds_value_single:
                    iter_resolve = resolver.iter_resolve_single
                    resolve = resolver.resolve_single
                    break
                
                if layout_flags.holds_value_multiple:
                    iter_resolve = resolver.iter_resolve_multiple
                    resolve = resolver.resolve_multiple
                    break
            
            iter_resolve = None
            resolve = None
            break
        
        self = PreinstancedBase.__new__(cls, value, name)
        self.layout_flags = layout_flags
        self.interaction_metadata_type = interaction_metadata_type
        self.iter_resolve = iter_resolve
        self.metadata_type = metadata_type
        self.resolve = resolve
        self.resolver = resolver
        return self
    
    
    none = P(
        0,
        'none',
        layout_flags = ComponentTypeLayoutFlag(),
    )
    
    row = P(
        1,
        'row',
        interaction_metadata_type = InteractionComponentMetadataRow,
        layout_flags = ComponentTypeLayoutFlag().update_by_keys(
            allowed_in_message = True,
            allowed_in_form = True,
            top_level = True,
            nestable_into_container = True,
            version_1 = True,
        ),
        metadata_type = ComponentMetadataRow,
    )
    
    button = P(
        2,
        'button',
        interaction_metadata_type = InteractionComponentMetadataButton,
        layout_flags = ComponentTypeLayoutFlag().update_by_keys(
            allowed_in_message = True,
            nestable_into_row = True,
            section_thumbnail = True,
            version_1 = True,
        ),
        metadata_type = ComponentMetadataButton,
    )
    
    string_select = P(
        3,
        'string select',
        interaction_metadata_type = InteractionComponentMetadataStringSelect,
        layout_flags = ComponentTypeLayoutFlag().update_by_keys(
            allowed_in_message = True,
            allowed_in_form = True,
            nestable_into_row = True,
            version_1 = True,
            nestable_into_label = True,
            holds_value_multiple = True,
        ),
        metadata_type = ComponentMetadataStringSelect,
        resolver = RESOLVER_STRING,
    )
    
    text_input = P(
        4,
        'text input',
        interaction_metadata_type = InteractionComponentMetadataTextInput,
        layout_flags = ComponentTypeLayoutFlag().update_by_keys(
            allowed_in_form = True,
            nestable_into_row = True,
            version_1 = True,
            nestable_into_label = True,
            holds_value_single = True,
        ),
        metadata_type = ComponentMetadataTextInput,
        resolver = RESOLVER_STRING,
    )
    
    user_select = P(
        5,
        'user select',
        interaction_metadata_type = InteractionComponentMetadataUserSelect,
        layout_flags = ComponentTypeLayoutFlag().update_by_keys(
            allowed_in_message = True,
            allowed_in_form = True,
            nestable_into_row = True,
            version_1 = True,
            nestable_into_label = True,
            holds_value_multiple = True,
        ),
        metadata_type = ComponentMetadataUserSelect,
        resolver = RESOLVER_USER,
    )
    
    role_select = P(
        6,
        'role select',
        interaction_metadata_type = InteractionComponentMetadataRoleSelect,
        layout_flags = ComponentTypeLayoutFlag().update_by_keys(
            allowed_in_message = True,
            allowed_in_form = True,
            nestable_into_row = True,
            version_1 = True,
            nestable_into_label = True,
            holds_value_multiple = True,
        ),
        metadata_type = ComponentMetadataRoleSelect,
        resolver = RESOLVER_ROLE,
    )
    
    mentionable_select = P(
        7,
        'mentionable select',
        interaction_metadata_type = InteractionComponentMetadataMentionableSelect,
        layout_flags = ComponentTypeLayoutFlag().update_by_keys(
            allowed_in_message = True,
            allowed_in_form = True,
            nestable_into_row = True,
            version_1 = True,
            nestable_into_label = True,
            holds_value_multiple = True,
        ),
        metadata_type = ComponentMetadataMentionableSelect,
        resolver = RESOLVER_MENTIONABLE,
    )
    
    channel_select = P(
        8,
        'channel select',
        interaction_metadata_type = InteractionComponentMetadataChannelSelect,
        layout_flags = ComponentTypeLayoutFlag().update_by_keys(
            allowed_in_message = True,
            allowed_in_form = True,
            nestable_into_row = True,
            version_1 = True,
            nestable_into_label = True,
            holds_value_multiple = True,
        ),
        metadata_type = ComponentMetadataChannelSelect,
        resolver = RESOLVER_CHANNEL,
    )
    
    section = P(
        9,
        'section',
        interaction_metadata_type = InteractionComponentMetadataSection,
        layout_flags = ComponentTypeLayoutFlag().update_by_keys(
            allowed_in_message = True,
            top_level = True,
            nestable_into_container = True,
            version_2 = True,
        ),
        metadata_type = ComponentMetadataSection,
    )
    
    text_display = P(
        10,
        'text display',
        layout_flags = ComponentTypeLayoutFlag().update_by_keys(
            allowed_in_message = True,
            allowed_in_form = True,
            top_level = True,
            nestable_into_container = True,
            nestable_into_section = True,
            version_2 = True,
        ),
        metadata_type = ComponentMetadataTextDisplay,
    )
    
    thumbnail_media = P(
        11,
        'thumbnail media',
        layout_flags = ComponentTypeLayoutFlag().update_by_keys(
            allowed_in_message = True,
            section_thumbnail = True,
            version_2 = True,
        ),
        metadata_type = ComponentMetadataThumbnailMedia,
    )
    
    media_gallery = P(
        12,
        'media gallery',
        layout_flags = ComponentTypeLayoutFlag().update_by_keys(
            allowed_in_message = True,
            top_level = True,
            nestable_into_container = True,
            version_2 = True,
        ),
        metadata_type = ComponentMetadataMediaGallery,
    )
    
    attachment_media = P(
        13,
        'attachment media',
        layout_flags = ComponentTypeLayoutFlag().update_by_keys(
            allowed_in_message = True,
            top_level = True,
            nestable_into_container = True,
            version_2 = True,
        ),
        metadata_type = ComponentMetadataAttachmentMedia,
    )
    
    separator = P(
        14,
        'separator',
        layout_flags = ComponentTypeLayoutFlag().update_by_keys(
            allowed_in_message = True,
            top_level = True,
            nestable_into_container = True,
            version_2 = True,
        ),
        metadata_type = ComponentMetadataSeparator,
    )
    
    # 15 ???
    
    # 16 ???
    
    container = P(
        17,
        'container',
        interaction_metadata_type = InteractionComponentMetadataContainer,
        layout_flags = ComponentTypeLayoutFlag().update_by_keys(
            allowed_in_message = True,
            top_level = True,
            version_2 = True,
        ),
        metadata_type = ComponentMetadataContainer,
    )
    
    label = P(
        18,
        'label',
        interaction_metadata_type = InteractionComponentMetadataLabel,
        layout_flags = ComponentTypeLayoutFlag().update_by_keys(
            allowed_in_form = True,
            top_level = True,
            version_2 = True,
        ),
        metadata_type = ComponentMetadataLabel,
    )
    
    attachment_input = P(
        19,
        'attachment input',
        interaction_metadata_type = InteractionComponentMetadataAttachmentInput,
        layout_flags = ComponentTypeLayoutFlag().update_by_keys(
            allowed_in_form = True,
            version_2 = True,
            nestable_into_label = True,
            holds_value_multiple = True,
        ),
        metadata_type = ComponentMetadataAttachmentInput,
        resolver = RESOLVER_ATTACHMENT,
    )



__all__ = ('ActivityTimestamps',)

from scarletio import copy_docs

from ...utils import DATETIME_FORMAT_CODE

from ..activity_field_base import ActivityFieldBase

from .fields import parse_end, parse_start, put_end, put_start, validate_end, validate_start


class ActivityTimestamps(ActivityFieldBase):
    """
    Represents an activity's timestamp field.
    
    Attributes
    ----------
    end : `None | DateTime`
        When the activity ends.
    
    start : `None | DateTime`
       When the activity starts.
    """
    __slots__ = ('end', 'start',)
    
    def __new__(cls, *, end = ..., start = ...):
        """
        Creates a new activity timestamp with the given parameters.
        
        Parameters
        ----------
        end : `None | DateTime`, Optional (Keyword only)
            When the activity ends. 
        start : `None | DateTime`, Optional (Keyword only)
           When the activity starts.
       
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # end
        if end is ...:
            end = None
        else:
            end = validate_end(end)
        
        # start
        if start is ...:
            start = None
        else:
            start = validate_start(start)
        
        # Construct
        self = object.__new__(cls)
        self.start = start
        self.end = end
        return self
    
    
    @classmethod
    @copy_docs(ActivityFieldBase.from_data)
    def from_data(cls, data):
        self = object.__new__(cls)
        self.start = parse_start(data)
        self.end = parse_end(data)
        return self
    
    
    @copy_docs(ActivityFieldBase.to_data)
    def to_data(self, *, defaults = False):
        data = {}
        put_end(self.end, data, defaults)
        put_start(self.start, data, defaults)
        return data
    
    
    @copy_docs(ActivityFieldBase.__repr__)
    def __repr__(self):
        repr_parts = [
            '<',
            type(self).__name__,
        ]
        
        start = self.start
        if (start is not None):
            repr_parts.append(' start = ')
            repr_parts.append(format(start, DATETIME_FORMAT_CODE))
            field_added = True
        else:
            field_added = False
        
        end = self.end
        if (end is not None):
            if field_added:
                repr_parts.append(',')
            
            repr_parts.append(' end = ')
            repr_parts.append(format(end, DATETIME_FORMAT_CODE))
        
        repr_parts.append('>')
        
        return ''.join(repr_parts)
    
    
    @copy_docs(ActivityFieldBase.__eq__)
    def __eq__(self, other):
        if type(self) is not type(other):
            return NotImplemented
        
        if self.end != other.end:
            return False
        
        if self.start != other.start:
            return False
        
        return True
    
    
    @copy_docs(ActivityFieldBase.__hash__)
    def __hash__(self):
        hash_value = 0
        
        end = self.end
        if (end is not None):
            hash_value ^= hash(end)
            hash_value ^= (1 << 0)
        
        start = self.start
        if (start is not None):
            hash_value ^= hash(start)
            hash_value ^= (1 << 4)
        
        return hash_value


    @copy_docs(ActivityFieldBase.__bool__)
    def __bool__(self):
        end = self.end
        if (end is not None):
            return True
        
        start = self.start
        if (start is not None):
            return True
        
        return False
    
    
    @copy_docs(ActivityFieldBase.copy)
    def copy(self):
        new  = object.__new__(type(self))
        new.end = self.end
        new.start = self.start
        return new
    
    
    def copy_with(self, *, end = ..., start = ...):
        """
        Copies the activity timestamp with the given fields.
        
        Parameters
        ----------
        end : `None | DateTime`, Optional (Keyword only)
            When the activity ends. 
        start : `None | DateTime`, Optional (Keyword only)
           When the activity starts.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # end
        if end is ...:
            end = self.end
        else:
            end = validate_end(end)
        
        # start
        if start is ...:
            start = self.start
        else:
            start = validate_start(start)
        
        # Construct
        new = object.__new__(type(self))
        new.start = start
        new.end = end
        return new



__all__ = ('ApplicationCommandOptionChoice',)

from scarletio import RichAttributeErrorBaseType

from ...localization.helpers import get_localized_length
from ...localization.utils import hash_locale_dictionary

from ..helpers import with_translation

from .fields import (
    parse_name, parse_name_localizations, parse_value, put_name, put_name_localizations, put_value,
    validate_name, validate_name_localizations, validate_value
)


class ApplicationCommandOptionChoice(RichAttributeErrorBaseType):
    """
    A choice of a ``ApplicationCommandOption``.
    
    Attributes
    ----------
    name : `str`
        The choice's name. It's length can be in range [1:100].
    name_localizations : ``None | dict<Locale, str>``
        Localized names of the choice.
    value : `str`, `int`, `float`
        The choice's value.
    """
    __slots__ = ('name', 'name_localizations', 'value')
    
    def __new__(cls, name, value = ..., *, name_localizations = ...):
        """
        Creates a new application command option choice with the given parameters.
        
        Parameters
        ----------
        name : `str`, `Enum`
            The choice's name. It's length can be in range [1:100].
        
        value : `None`, `str`, `int`, `float`, `Enum`, Optional
            The choice's value.
            
            Defaults to `name` parameter if not given.
        
        name_localizations : ``None | dict<str | Locale, str> | (list | set | tuple<(str | Locale, str>)`` \
                , Optional (Keyword only)
            Localized names of the choice.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's length is incorrect.
        """
        if value is ...:
            value = name
        
        name = validate_name(name)
        value = validate_value(value)
        
        if name_localizations is ...:
            name_localizations = None
        else:
            name_localizations = validate_name_localizations(name_localizations)
        
        self = object.__new__(cls)
        self.name = name
        self.name_localizations = name_localizations
        self.value = value
        return self
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new application command option choice from the received data.
        
        Parameters
        ----------
        data : `dict<str, object>`
            The received application command option choice data.
        
        Returns
        -------
        self : `instance<cls>`
            The created choice.
        """
        self = object.__new__(cls)
        self.name = parse_name(data)
        self.name_localizations = parse_name_localizations(data)
        self.value = parse_value(data)
        return self
    
    
    def to_data(self, *, defaults = False):
        """
        Converts the application command option choice to a json serializable object.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether fields with their default value should be included as well.
        
        Returns
        -------
        data : `dict<str, object>`
        """
        data = {}
        put_name(self.name, data, defaults)
        put_name_localizations(self.name_localizations, data, defaults)
        put_value(self.value, data, defaults)
        return data
    
    
    def __repr__(self):
        """Returns the application command option choice's representation."""
        repr_parts = ['<', self.__class__.__name__]
        
        # name
        repr_parts.append(' name = ')
        repr_parts.append(repr(self.name))
        
        # value
        repr_parts.append(', value = ')
        repr_parts.append(repr(self.value))
        
        # Extra fields: `.name_localizations`
        
        # name_localizations
        name_localizations = self.name_localizations
        if (name_localizations is not None):
            repr_parts.append(', name_localizations = ')
            repr_parts.append(repr(name_localizations))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __eq__(self, other):
        """Returns whether the two choices are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        # name
        if self.name != other.name:
            return False
        
        # name_localizations
        if self.name_localizations != other.name_localizations:
            return False
        
        # value
        if self.value != other.value:
            return False
        
        return True
    
    
    def __len__(self):
        """Returns the application command option choice's length."""
        length = 0
        
        # name & name_localizations
        length += get_localized_length(self.name, self.name_localizations)
        
        # value
        value = self.value
        if isinstance(value, str):
            length += len(value)
        
        return length
    
    
    def __hash__(self):
        """Returns the application command option choice's representation."""
        hash_value = 0
        
        # name
        hash_value ^= hash(self.name)
        
        # name_localizations
        name_localizations = self.name_localizations
        if (name_localizations is not None):
            hash_value ^= hash_locale_dictionary(name_localizations)
        
        # value
        # Do not hash `.value` of equals to `.name`
        value = self.value
        if (not isinstance(value, str)) or (value != self.name):
            hash_value ^= hash(value)
        
        return hash_value
    
    
    def copy(self):
        """
        Copies the application command option choice.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        
        # name
        new.name = self.name
        
        # name_localizations
        name_localizations = self.name_localizations
        if (name_localizations is not None):
            name_localizations = name_localizations.copy()
        new.name_localizations = name_localizations
        
        # value
        new.value = self.value
        
        return new
    
    
    def copy_with(self, *, name = ..., value = ..., name_localizations = ...):
        """
        Copies the application command option choice with the given fields.
        
        > Not like ``.__new__``, ``.copy_with`` will not default `value` to `name` if `value` is not provided.
        
        Parameters
        ----------
        name : `str`, `Enum`, Optional (Keyword only)
            The choice's name. It's length can be in range [1:100].
        
        name_localizations : ``None | dict<str | Locale, str> | (list | set | tuple<(str | Locale, str>)`` \
                , Optional (Keyword only)
            Localized names of the choice.
        
        value : `None`, `str`, `int`, `float`, `Enum`, Optional (Keyword only)
            The choice's value.
            
            Defaults to `name` parameter if not given.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's length is incorrect.
        """
        # name
        if name is ...:
            name = self.name
        else:
            name = validate_name(name)
        
        # name_localizations
        if name_localizations is ...:
            name_localizations = self.name_localizations
            if (name_localizations is not None):
                name_localizations = name_localizations.copy()
        else:
            name_localizations = validate_name_localizations(name_localizations)
        
        # value
        if value is ...:
            value = self.value
        else:
            value = validate_value(value)
        new = object.__new__(type(self))
        
        # Construct
        new.name = name
        new.name_localizations = name_localizations
        new.value = value
        return new

    
    def with_translation(self, translation_table, replace = False):
        """
        Returns a new application command option choice with the given translation table applied.
        
        Parameters
        ----------
        translation_table : `None`, `dict` of ((``Locale``, `str`), \
                (`None`, `dict` (`str`, (`None`, `str`)) items)) items
            Translation table to pull localization. from.
        replace : `bool` = `False`, Optional
            Whether actual translation should be replaced.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        if translation_table is None:
            return self
        
        new = self.copy()
        
        # name_localizations
        new.name_localizations = with_translation(
            new.name,
            new.name_localizations,
            translation_table,
            replace,
        )
        
        return new



from .attachment import *
from .message import *
from .message_activity import *
from .message_application import *
from .message_builder import *
from .message_call import *
from .message_interaction import *
from .message_pin import *
from .message_role_subscription import *
from .message_search_query import *
from .message_search_response import *
from .message_snapshot import *
from .poll_change import *
from .poll_update import *
from .shared_client_theme import *
from .voice_attachment import *


__all__ = (
    *attachment.__all__,
    *message.__all__,
    *message_activity.__all__,
    *message_application.__all__,
    *message_builder.__all__,
    *message_call.__all__,
    *message_interaction.__all__,
    *message_pin.__all__,
    *message_role_subscription.__all__,
    *message_search_query.__all__,
    *message_search_response.__all__,
    *message_snapshot.__all__,
    *poll_change.__all__,
    *poll_update.__all__,
    *shared_client_theme.__all__,
    *voice_attachment.__all__,
)



__all__ = ('ApplicationCommandOptionMetadataNumeric',)

from scarletio import copy_docs

from .fields import (
    parse_max_value, parse_min_value, put_max_value, put_min_value, validate_max_value_postprocessed,
    validate_min_value_postprocessed
)
from .primitive import ApplicationCommandOptionMetadataPrimitive


class ApplicationCommandOptionMetadataNumeric(ApplicationCommandOptionMetadataPrimitive):
    """
    Base type for numeric parameter application command option metadatas.
    
    Attributes
    ----------
    autocomplete : `bool`
        Whether the option supports auto completion.
    
    choices : `None`, `tuple` of ``ApplicationCommandOptionChoice``
        Choices for the user to pick from.
    
    max_value : `None`, `float`, `int`
        The maximal value permitted for this option.
    
    min_value : `None`, `float`, ``int`
        The minimum value permitted for this option.
    
    required : `bool`
        Whether the parameter is required. Defaults to `False`.
    """
    __slots__ = ('max_value', 'min_value')
    
    
    def __new__(cls, *, autocomplete = ..., choices = ..., max_value = ..., min_value = ..., required = ...):
        """
        Creates a new numeric application command option metadata with the given parameters.
        
        Parameters
        ----------
        autocomplete : `bool`, Optional (Keyword only)
            Whether the option supports auto completion.
        
        choices : `None`, `iterable` of ``ApplicationCommandOptionChoice``, Optional (Keyword only)
            Choices for the user to pick from.
        
        max_value : `None`, `float`, `int`, Optional (Keyword only)
            The maximal value permitted for this option.
        
        min_value : `None`, `float`, ``int`, Optional (Keyword only)
            The minimum value permitted for this option.
        
        required : `bool`, Optional (Keyword only)
            Whether the parameter is required.
        
        Raises
        ------
        TypeError
            - If a parameter of incorrect type given.
        ValueError
            - If a parameter of incorrect value given.
        """
        # max_value
        if max_value is ...:
            max_value = None
        else:
            max_value = validate_max_value_postprocessed(max_value, cls.TYPE)
        
        # min_value
        if min_value is ...:
            min_value = None
        else:
            min_value = validate_min_value_postprocessed(min_value, cls.TYPE)
        
        # Construct
        new = ApplicationCommandOptionMetadataPrimitive.__new__(
            cls,
            autocomplete = autocomplete,
            choices = choices,
            required = required,
        )
        new.max_value = max_value
        new.min_value = min_value
        return new
    
    
    @classmethod
    @copy_docs(ApplicationCommandOptionMetadataPrimitive.from_keyword_parameters)
    def from_keyword_parameters(cls, keyword_parameters):
        return cls(
            autocomplete = keyword_parameters.pop('autocomplete', ...),
            choices = keyword_parameters.pop('choices', ...),
            max_value = keyword_parameters.pop('max_value', ...),
            min_value = keyword_parameters.pop('min_value', ...),
            required = keyword_parameters.pop('required', ...),
        )
    
    
    @classmethod
    @copy_docs(ApplicationCommandOptionMetadataPrimitive.from_data)
    def from_data(cls, data):
        self = super(ApplicationCommandOptionMetadataNumeric, cls).from_data(data)
        self.max_value = parse_max_value(data)
        self.min_value = parse_min_value(data)
        return self
    
    
    @copy_docs(ApplicationCommandOptionMetadataPrimitive.to_data)
    def to_data(self, *, defaults = False):
        data = ApplicationCommandOptionMetadataPrimitive.to_data(self, defaults = defaults)
        put_max_value(self.max_value, data, defaults)
        put_min_value(self.min_value, data, defaults)
        return data
    
    
    @copy_docs(ApplicationCommandOptionMetadataPrimitive._add_type_specific_repr_fields)
    def _add_type_specific_repr_fields(self, repr_parts):
        ApplicationCommandOptionMetadataPrimitive._add_type_specific_repr_fields(self, repr_parts)
        
        repr_parts.append(', max_value = ')
        repr_parts.append(repr(self.max_value))
        
        repr_parts.append(', min_value = ')
        repr_parts.append(repr(self.min_value))
    
    
    @copy_docs(ApplicationCommandOptionMetadataPrimitive._is_equal_same_type)
    def _is_equal_same_type(self, other):
        if not ApplicationCommandOptionMetadataPrimitive._is_equal_same_type(self, other):
            return False
        
        if self.max_value != other.max_value:
            return False
        
        if self.min_value != other.min_value:
            return False
        
        return True
    
    
    @copy_docs(ApplicationCommandOptionMetadataPrimitive.__hash__)
    def __hash__(self):
        hash_value = ApplicationCommandOptionMetadataPrimitive.__hash__(self)
        
        # max_value
        hash_value ^= hash(self.max_value) << 6
        
        # min_value
        hash_value ^= hash(self.min_value) << 10
        
        return hash_value
    
    
    @copy_docs(ApplicationCommandOptionMetadataPrimitive.copy)
    def copy(self):
        new = ApplicationCommandOptionMetadataPrimitive.copy(self)
        new.max_value = self.max_value
        
        # min_value
        new.min_value = self.min_value
        
        return new
    
    
    def copy_with(self, *, autocomplete = ..., choices = ..., max_value = ..., min_value = ..., required = ...):
        """
        Copies the numeric application command option metadata with the given fields.
        
        Parameters
        ----------
        autocomplete : `bool`, Optional (Keyword only)
            Whether the option supports auto completion.
        
        choices : `None`, `iterable` of ``ApplicationCommandOptionChoice``, Optional (Keyword only)
            Choices for the user to pick from.
        
        max_value : `None`, `float`, `int`, Optional (Keyword only)
            The maximal value permitted for this option.
        
        min_value : `None`, `float`, ``int`, Optional (Keyword only)
            The minimum value permitted for this option.
        
        required : `bool`, Optional (Keyword only)
            Whether the parameter is required.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter of incorrect type given.
        ValueError
            - If a parameter of incorrect value given.
        """
        # max_value
        if max_value is ...:
            max_value = self.max_value
        else:
            max_value = validate_max_value_postprocessed(max_value, type(self).TYPE)
        
        # min_value
        if min_value is ...:
            min_value = self.min_value
        else:
            min_value = validate_min_value_postprocessed(min_value, type(self).TYPE)
        
        # Construct
        new = ApplicationCommandOptionMetadataPrimitive.copy_with(
            self,
            autocomplete = autocomplete,
            choices = choices,
            required = required,
        )
        new.max_value = max_value
        new.min_value = min_value
        return new
    
    
    @copy_docs(ApplicationCommandOptionMetadataPrimitive.copy_with_keyword_parameters)
    def copy_with_keyword_parameters(self, keyword_parameters):
        return self.copy_with(
            autocomplete = keyword_parameters.pop('autocomplete', ...),
            choices = keyword_parameters.pop('choices', ...),
            max_value = keyword_parameters.pop('max_value', ...),
            min_value = keyword_parameters.pop('min_value', ...),
            required = keyword_parameters.pop('required', ...),
        )
